import time

from datetime import timedelta
from collections import defaultdict
from typing import DefaultDict, List, Optional, Sequence, Set, Tuple

from glycresoft.structure.scan import ScanInformation
from glycresoft.trace.extract import ChromatogramExtractor

from glypy.composition import formula

from glycresoft.task import TaskBase
from glycresoft.chromatogram_tree import (
    ChromatogramFilter,
    DisjointChromatogramSet,
    Chromatogram
)

from glycresoft.database.builder.glycan.migrate import (
    GlycanHypothesisMigrator)

from glycresoft.database.builder.glycopeptide.migrate import (
    GlycopeptideHypothesisMigrator,
    Glycopeptide,
    Peptide,
    Protein,
    GlycanCombination)

from glycresoft.serialize import (
    AnalysisTypeEnum,
    ChromatogramSolutionMassShiftedToChromatogramSolution)

from glycresoft.tandem.glycopeptide.identified_structure import (
    IdentifiedGlycopeptide as MemoryIdentifiedGlycopeptide)
from glycresoft.scoring.chromatogram_solution import ChromatogramSolution

from glycresoft.serialize.utils import toggle_indices
from glycresoft.serialize.hypothesis import GlycanComposition
from glycresoft.serialize import DatabaseBoundOperation
from glycresoft.serialize.serializer import AnalysisSerializer

from glycresoft.serialize.spectrum import (
    SampleRun,
    MSScan,
    DeconvolutedPeak,
    PrecursorInformation)

from ms_deisotope.data_source import ChargeNotProvided
from ms_deisotope.output.common import SampleRun as MemorySampleRun

def slurp(session, model, ids):
    ids = sorted(ids)
    total = len(ids)
    last = 0
    step = 100
    results = []
    while last < total:
        results.extend(session.query(model).filter(
            model.id.in_(ids[last:last + step])))
        last += step
    return results


def fetch_scans_used_in_chromatogram(chromatogram_set: List[Chromatogram],
                                     extractor: ChromatogramExtractor) -> List[ScanInformation]:
    scan_ids = set()
    for chroma in chromatogram_set:
        scan_ids.update(chroma.scan_ids)

        tandem_solutions = getattr(chroma, "tandem_solutions", [])
        for tsm in tandem_solutions:
            scan_ids.add(tsm.scan.id)
            scan_ids.add(tsm.scan.precursor_information.precursor_scan_id)

    scans = []
    with extractor.toggle_peak_loading():
        for scan_id in scan_ids:
            scans.append(
                ScanInformation.from_scan(extractor.get_scan_header_by_id(scan_id))
            )
    return sorted(scans, key=lambda x: (x.ms_level, x.index))


def fetch_glycan_compositions_from_chromatograms(chromatogram_set: Sequence[ChromatogramSolution],
                                                 glycan_db: DatabaseBoundOperation) -> List[GlycanComposition]:
    ids = set()
    for chroma in chromatogram_set:
        if chroma.composition:
            ids.add(chroma.composition.id)
    glycan_compositions = []
    for glycan_id in ids:
        gc = glycan_db.query(GlycanComposition).get(glycan_id)
        glycan_compositions.append(gc)
    return glycan_compositions


def update_glycan_chromatogram_composition_ids(hypothesis_migration,
                                               glycan_chromatograms: Sequence[ChromatogramSolution]):
    mapping = dict()
    tandem_mapping = defaultdict(list)
    for chromatogram in glycan_chromatograms:
        if chromatogram.composition is None:
            continue
        mapping[chromatogram.composition.id] = chromatogram.composition
        for match in getattr(chromatogram, 'tandem_solutions', []):
            tandem_mapping[match.target.id].append(match.target)

    for key, value in mapping.items():
        value.id = hypothesis_migration.glycan_composition_id_map[value.id]
    for key, group in tandem_mapping.items():
        for value in group:
            value.id = hypothesis_migration.glycan_composition_id_map[key]


class SampleMigrator(DatabaseBoundOperation, TaskBase):
    def __init__(self, connection):
        DatabaseBoundOperation.__init__(self, connection)
        self.sample_run_id = None
        self.ms_scan_id_map = dict()
        self.peak_id_map = dict()

    def _migrate(self, obj):
        self.session.add(obj)
        self.session.flush()
        new_id = obj.id
        return new_id

    def commit(self):
        self.session.commit()

    def clear(self):
        self.peak_id_map.clear()
        self.ms_scan_id_map.clear()

    def migrate_sample_run(self, sample_run):
        new_sample_run = SampleRun(
            name=sample_run.name,
            uuid=sample_run.uuid,
            sample_type=sample_run.sample_type,
            completed=sample_run.completed
        )
        new_id = self._migrate(new_sample_run)
        self.sample_run_id = new_id

    def scan_id(self, ms_scan) -> str:
        if ms_scan is None:
            return None
        try:
            scan_id = ms_scan.scan_id
        except AttributeError:
            scan_id = ms_scan.id
        return scan_id

    def peak_id(self, peak, scan_id) -> Tuple[str, DeconvolutedPeak]:
        try:
            if peak.id:
                peak_id = (scan_id, peak.convert())
        except AttributeError:
            peak_id = (scan_id, peak)
        return peak_id

    def migrate_precursor_information(self, prec_info):
        try:
            # prec_id = self.scan_id(prec_info.precursor)
            prec_id = prec_info.precursor_scan_id
        except KeyError as err:
            prec_id = None
            self.log("Unable to locate precursor scan with ID %r" % (err , ))
        # prod_id = self.scan_id(prec_info.product)
        prod_id = prec_info.product_scan_id
        charge = prec_info.charge
        if charge == ChargeNotProvided:
            charge = 0
        new_info = PrecursorInformation(
            sample_run_id=self.sample_run_id,
            # precursor may be missing
            precursor_id=self.ms_scan_id_map.get(prec_id),
            product_id=self.ms_scan_id_map[prod_id],
            neutral_mass=prec_info.neutral_mass,
            charge=charge,
            intensity=prec_info.intensity)
        return self._migrate(new_info)

    def migrate_ms_scan(self, ms_scan):
        if ms_scan is None:
            return
        scan_id = self.scan_id(ms_scan)
        new_scan = MSScan._serialize_scan(ms_scan, self.sample_run_id)
        try:
            new_scan.info.update(dict(ms_scan.info))
        except AttributeError:
            pass
        new_id = self._migrate(new_scan)
        self.ms_scan_id_map[scan_id] = new_id
        if ms_scan.precursor_information:
            self.migrate_precursor_information(
                ms_scan.precursor_information)

    def migrate_peak(self, peak, scan_id):
        new_scan_id = self.ms_scan_id_map[scan_id]
        new_peak = DeconvolutedPeak.serialize(peak)
        new_peak.scan_id = new_scan_id
        new_id = self._migrate(new_peak)
        peak_id = self.peak_id(peak, scan_id)
        self.peak_id_map[peak_id] = new_id


class AnalysisMigrationBase(DatabaseBoundOperation, TaskBase):
    sample_run: MemorySampleRun
    chromatogram_extractor: ChromatogramExtractor

    _seed_analysis_name: str
    _analysis_serializer: Optional[AnalysisSerializer]
    _sample_migrator: Optional[SampleMigrator]

    def __init__(self, connection: str, analysis_name: str,
                 sample_run: MemorySampleRun,
                 chromatogram_extractor: ChromatogramExtractor):
        DatabaseBoundOperation.__init__(self, connection)

        self.sample_run = sample_run
        self.chromatogram_extractor = chromatogram_extractor

        self._seed_analysis_name = analysis_name
        self._analysis_serializer = None
        self._sample_migrator = None

    @property
    def analysis(self):
        return self._analysis_serializer.analysis

    def set_analysis_type(self, type_string: str):
        self._analysis_serializer.analysis.analysis_type = type_string
        self._analysis_serializer.session.add(self.analysis)
        self._analysis_serializer.session.commit()

    def set_parameters(self, parameters: dict):
        self._analysis_serializer.analysis.parameters = parameters
        self._analysis_serializer.session.add(self.analysis)
        self._analysis_serializer.commit()

    def fetch_peaks(self, chromatogram_set: Sequence[Chromatogram]) -> DefaultDict[str, Set[DeconvolutedPeak]]:
        scan_peak_map = defaultdict(set)
        for chroma in chromatogram_set:
            if chroma is None:
                continue
            for node in chroma:
                for peak in node.peaks:
                    scan_peak_map[node.scan_id].add(peak)
        return scan_peak_map

    def fetch_scans(self) -> List[ScanInformation]:
        raise NotImplementedError()

    def migrate_sample(self):
        self.log("... Loading Scans")
        scans = self.fetch_scans()
        self._sample_migrator = SampleMigrator(self._original_connection)
        self._sample_migrator.migrate_sample_run(self.sample_run)

        self.log("... Migrating Scans")
        for scan in sorted(scans, key=lambda x: x.ms_level):
            self._sample_migrator.migrate_ms_scan(scan)

        scan_peak_map = self.fetch_peaks(self.chromatogram_set)
        self.log("... Migrating Peaks")
        for scan_id, peaks in scan_peak_map.items():
            for peak in peaks:
                self._sample_migrator.migrate_peak(peak, scan_id)

        self._sample_migrator.commit()

    def migrate_hypothesis(self):
        raise NotImplementedError()

    def create_analysis(self):
        raise NotImplementedError()

    def run(self):
        self.log("Migrating Hypothesis")
        self.migrate_hypothesis()
        self.log("Migrating Sample Run")
        self.migrate_sample()
        self.log("Creating Analysis Record")
        self.create_analysis()


class GlycanCompositionChromatogramAnalysisSerializer(AnalysisMigrationBase):
    _glycan_hypothesis_migrator: Optional[GlycanHypothesisMigrator]
    glycan_db: DatabaseBoundOperation
    chromatogram_set: ChromatogramFilter

    def __init__(self, connection, analysis_name, sample_run,
                 chromatogram_set, glycan_db,
                 chromatogram_extractor):
        AnalysisMigrationBase.__init__(
            self, connection, analysis_name, sample_run,
            chromatogram_extractor)
        self._glycan_hypothesis_migrator = None

        self.glycan_db = glycan_db
        self.chromatogram_set = ChromatogramFilter(chromatogram_set)
        self._index_chromatogram_set()

    def _index_chromatogram_set(self):
        for i, chroma in enumerate(self.chromatogram_set):
            chroma.id = i

    @property
    def analysis(self):
        return self._analysis_serializer.analysis

    def fetch_scans(self):
        scans = fetch_scans_used_in_chromatogram(
            self.chromatogram_set, self.chromatogram_extractor)
        return scans

    def migrate_hypothesis(self):
        self._glycan_hypothesis_migrator = GlycanHypothesisMigrator(
            self._original_connection)
        self._glycan_hypothesis_migrator.migrate_hypothesis(
            self.glycan_db.hypothesis)
        for glycan_composition in fetch_glycan_compositions_from_chromatograms(
                self.chromatogram_set, self.glycan_db):
            self._glycan_hypothesis_migrator.migrate_glycan_composition(
                glycan_composition)
        self._glycan_hypothesis_migrator.commit()

    def locate_mass_shift_reference_solution(self, chromatogram):
        if chromatogram.used_as_mass_shift:
            for key, mass_shift in chromatogram.used_as_mass_shift:
                disjoint_set_for_key = self.chromatogram_set.key_map[key]
                if not isinstance(disjoint_set_for_key, DisjointChromatogramSet):
                    disjoint_set_for_key = DisjointChromatogramSet(disjoint_set_for_key)
                hit = disjoint_set_for_key.find_overlap(chromatogram)
                if hit is not None:
                    yield hit, mass_shift

    def _get_solution_db_id_by_reference(self, ref_id):
        return self._analysis_serializer._chromatogram_solution_id_map[ref_id]

    def _get_mass_shift_id(self, mass_shift):
        return self._analysis_serializer._mass_shift_cache.serialize(mass_shift).id

    def express_mass_shift_relation(self, chromatogram: Chromatogram):
        try:
            source_id = self._get_solution_db_id_by_reference(chromatogram.id)
        except KeyError as e:
            self.error(
                    f"Chromatogram not registered: {e} -> {chromatogram}")
        seen = set()
        for related_solution, mass_shift in self.locate_mass_shift_reference_solution(
                chromatogram):
            if related_solution.id in seen:
                continue
            else:
                seen.add(related_solution.id)
            try:
                referenced_id = self._get_solution_db_id_by_reference(related_solution.id)
                mass_shift_id = self._get_mass_shift_id(mass_shift)
                self._analysis_serializer.session.execute(
                    ChromatogramSolutionMassShiftedToChromatogramSolution.__table__.insert(),
                    {
                        "mass_shifted_solution_id": source_id,
                        "owning_solution_id": referenced_id,
                        "mass_shift_id": mass_shift_id
                    })
            except KeyError as e:
                self.error(
                    f"Failed to express mass shift relation: {e} -> {mass_shift} @ {chromatogram}")
                continue

    def create_analysis(self):
        self._analysis_serializer = AnalysisSerializer(
            self._original_connection,
            analysis_name=self._seed_analysis_name,
            sample_run_id=self._sample_migrator.sample_run_id)

        self._analysis_serializer.set_analysis_type(AnalysisTypeEnum.glycan_lc_ms.name)

        update_glycan_chromatogram_composition_ids(
            self._glycan_hypothesis_migrator,
            self.chromatogram_set)

        self._analysis_serializer.set_peak_lookup_table(self._sample_migrator.peak_id_map)

        n = len(self.chromatogram_set)
        i = 0

        for chroma in self.chromatogram_set:
            i += 1
            if i % 100 == 0:
                self.log("%0.2f%% of Chromatograms Saved (%d/%d)" % (
                    i * 100. / n, i, n))
            if chroma.composition is not None:
                self._analysis_serializer.save_glycan_composition_chromatogram_solution(
                    chroma)
            else:
                self._analysis_serializer.save_unidentified_chromatogram_solution(
                    chroma)

        for chroma in self.chromatogram_set:
            if chroma.chromatogram.used_as_mass_shift:
                self.express_mass_shift_relation(chroma)

        self._analysis_serializer.commit()


class GlycopeptideMSMSAnalysisSerializer(AnalysisMigrationBase):


    _glycopeptide_hypothesis_migrator: GlycopeptideHypothesisMigrator
    _glycopeptide_ids: Set[int]

    _identified_glycopeptide_set: List[MemoryIdentifiedGlycopeptide]
    _unassigned_chromatogram_set: List[Chromatogram]


    def __init__(self, connection, analysis_name, sample_run,
                 identified_glycopeptide_set, unassigned_chromatogram_set,
                 glycopeptide_db, chromatogram_extractor):
        AnalysisMigrationBase.__init__(
            self, connection, analysis_name, sample_run, chromatogram_extractor)

        self._glycopeptide_hypothesis_migrator = None

        self.glycopeptide_db = glycopeptide_db
        self._identified_glycopeptide_set = identified_glycopeptide_set

        self._unassigned_chromatogram_set = unassigned_chromatogram_set
        self._glycopeptide_ids = None

        self.aggregate_glycopeptide_ids()

    def aggregate_glycopeptide_ids(self):
        """Accumulate a set of all unique structure IDs over all spectrum matches.

        Returns
        -------
        set
        """
        aggregate = set()
        for gp in self._identified_glycopeptide_set:
            for solution_set in gp.spectrum_matches:
                for match in solution_set:
                    aggregate.add(match.target.id)
        self._glycopeptide_ids = aggregate

    def update_glycopeptide_ids(self):
        '''Build a mapping from searched hypothesis structure ID to saved subset structure
        ID, ensuring that each object has its new ID assigned exactly once.

        The use of :func:`id` here is to ensure that exactness.
        '''
        hypothesis_migration = self._glycopeptide_hypothesis_migrator
        identified_glycopeptide_set = self._identified_glycopeptide_set
        mapping = defaultdict(dict)
        for glycopeptide in identified_glycopeptide_set:
            mapping[glycopeptide.structure.id][id(
                glycopeptide.structure)] = glycopeptide.structure
            for solution_set in glycopeptide.spectrum_matches:
                for match in solution_set:
                    mapping[match.target.id][id(match.target)] = match.target
        for key, instances in mapping.items():
            for value in instances.values():
                try:
                    value.id = hypothesis_migration.glycopeptide_id_map[value.id]
                except KeyError:
                    self.log(f"Could not find a mapping from ID {value.id} for {value} "
                              "in the new keyspace, reconstructing to recover...")
                    value.id = self._migrate_single_glycopeptide(value)

    @property
    def chromatogram_set(self):
        return [
            gp.chromatogram for gp in self._identified_glycopeptide_set
        ] + list(self._unassigned_chromatogram_set)

    def migrate_hypothesis(self):
        start = time.time()
        self._glycopeptide_hypothesis_migrator = GlycopeptideHypothesisMigrator(
            self._original_connection)
        self._glycopeptide_hypothesis_migrator.migrate_hypothesis(
            self.glycopeptide_db.hypothesis)

        self.log("... Migrating Glycans")
        glycan_compositions, glycan_combinations = self.fetch_glycan_compositions(
            self._glycopeptide_ids)

        for i, gc in enumerate(glycan_compositions):
            if i % 1000 == 0 and i:
                self.log("...... Migrating Glycan %d" % (i, ))
            self._glycopeptide_hypothesis_migrator.migrate_glycan_composition(gc)
        for i, gc in enumerate(glycan_combinations):
            if i % 1000 == 0 and i:
                self.log("...... Migrating Glycan Combination %d" % (i, ))
            self._glycopeptide_hypothesis_migrator.migrate_glycan_combination(gc)

        self.log("... Loading Proteins and Peptides from Glycopeptides")
        peptides = self.fetch_peptides(self._glycopeptide_ids)
        proteins = self.fetch_proteins(peptides)
        self.log("... Migrating Proteins (%d)" % (len(proteins), ))
        n = len(proteins)
        index_toggler = toggle_indices(self._glycopeptide_hypothesis_migrator.session, Protein)
        index_toggler.drop()
        for i, protein in enumerate(proteins):
            if i % 5000 == 0 and i:
                self.log("...... Migrating Protein %d/%d (%0.2f%%) %s" % (i, n, i * 100.0 / n, protein.name))
            self._glycopeptide_hypothesis_migrator.migrate_protein(protein)
        proteins = []
        index_toggler.create()
        index_toggler = toggle_indices(
            self._glycopeptide_hypothesis_migrator.session, Peptide)
        index_toggler.drop()
        n = len(peptides)
        self.log("... Migrating Peptides (%d)" % (n, ))
        self._glycopeptide_hypothesis_migrator.migrate_peptides_bulk(peptides)
        index_toggler.create()
        peptides = []

        glycopeptides = self.fetch_glycopeptides(self._glycopeptide_ids)

        index_toggler = toggle_indices(self._glycopeptide_hypothesis_migrator.session, Glycopeptide)
        index_toggler.drop()
        n = len(glycopeptides)
        self.log("... Migrating Glycopeptides (%d)" % (n, ))
        self._glycopeptide_hypothesis_migrator.migrate_glycopeptide_bulk(glycopeptides)
        index_toggler.create()
        self._glycopeptide_hypothesis_migrator.commit()
        elapsed = timedelta(seconds=time.time()) - timedelta(seconds=start)
        self.log(f"... Hypothesis migration time {elapsed}")

    def fetch_scans(self):
        scan_ids = set()
        precursor_ids = dict()
        with self.chromatogram_extractor.toggle_peak_loading():
            for glycopeptide in self._identified_glycopeptide_set:
                if glycopeptide.chromatogram is not None:
                    scan_ids.update(glycopeptide.chromatogram.scan_ids)
                for msms in glycopeptide.spectrum_matches:
                    scan_ids.add(msms.scan.id)
                    try:
                        scan_ids.add(precursor_ids[msms.scan.id])
                    except KeyError:
                        precursor_ids[msms.scan.id] = self.chromatogram_extractor.get_scan_header_by_id(
                            msms.scan.id).precursor_information.precursor_scan_id
                        scan_ids.add(precursor_ids[msms.scan.id])
            scans = []
            for scan_id in scan_ids:
                if scan_id is not None:
                    try:
                        scans.append(ScanInformation.from_scan(self.chromatogram_extractor.get_scan_header_by_id(scan_id)))
                    except KeyError:
                        self.log("Could not locate scan for ID %r" % (scan_id, ))
        return sorted(scans, key=lambda x: (x.ms_level, x.index))

    def _get_glycan_combination_for_glycopeptide(self, glycopeptide_id: int) -> int:
        gc_comb_id = self.glycopeptide_db.query(
            Glycopeptide.glycan_combination_id).filter(
            Glycopeptide.id == glycopeptide_id)
        return gc_comb_id[0]

    def fetch_glycan_compositions(self, glycopeptide_ids: Set[int]) -> Tuple[List[GlycanComposition],
                                                                             List[GlycanCombination]]:
        glycan_compositions = dict()
        glycan_combinations = dict()

        glycan_combination_ids = set()
        for glycopeptide_id in glycopeptide_ids:
            gc_comb_id = self._get_glycan_combination_for_glycopeptide(glycopeptide_id)
            glycan_combination_ids.add(gc_comb_id)

        for i in glycan_combination_ids:
            gc_comb = self.glycopeptide_db.query(GlycanCombination).get(i)
            glycan_combinations[gc_comb.id] = gc_comb

            for composition in gc_comb.components:
                glycan_compositions[composition.id] = composition

        return list(glycan_compositions.values()), list(glycan_combinations.values())

    def _get_peptide_id_for_glycopeptide(self, glycopeptide_id: int) -> int:
        db_glycopeptide = self.glycopeptide_db.query(Glycopeptide).get(glycopeptide_id)
        return db_glycopeptide.peptide_id

    def fetch_peptides(self, glycopeptide_ids: Set[int]) -> List[Peptide]:
        peptides = set()
        for glycopeptide_id in glycopeptide_ids:
            peptide_id = self._get_peptide_id_for_glycopeptide(glycopeptide_id)
            peptides.add(peptide_id)

        return slurp(self.glycopeptide_db, Peptide, peptides)

    def fetch_proteins(self, peptide_set: List[Peptide]) -> List[Protein]:
        proteins = set()
        for peptide in peptide_set:
            proteins.add(peptide.protein_id)
        return slurp(self.glycopeptide_db, Protein, proteins)

    def fetch_glycopeptides(self, glycopeptide_ids: Set[int]) -> List[Glycopeptide]:
        return slurp(self.glycopeptide_db, Glycopeptide, glycopeptide_ids)

    def _migrate_single_glycopeptide(self, glycopeptide: Glycopeptide) -> int:
        """
        Stupendously slow if used in bulk, intended for recovering from cases where
        a single glycopeptide somehow slipped through the cracks.
        """
        gp = self.glycopeptide_db.query(Glycopeptide).get(id)
        self._glycopeptide_hypothesis_migrator.migrate_glycopeptide(gp)
        return self._glycopeptide_hypothesis_migrator.glycopeptide_id_map[glycopeptide.id]

    def create_analysis(self):
        self._analysis_serializer = AnalysisSerializer(
            self._original_connection,
            analysis_name=self._seed_analysis_name,
            sample_run_id=self._sample_migrator.sample_run_id)

        self._analysis_serializer.set_analysis_type(
            AnalysisTypeEnum.glycopeptide_lc_msms.name)

        self._analysis_serializer.set_peak_lookup_table(
            self._sample_migrator.peak_id_map)

        # Patch the in-memory objects to be mapped through to the new database keyspace
        self.update_glycopeptide_ids()

        # Free up memory from the glycopeptide identity map
        self._glycopeptide_hypothesis_migrator.clear()
        self._analysis_serializer.save_glycopeptide_identification_set(
            self._identified_glycopeptide_set)

        for chroma in self._unassigned_chromatogram_set:
            self._analysis_serializer.save_unidentified_chromatogram_solution(
                chroma)

        self.log("Committing Analysis")
        self._analysis_serializer.commit()


class DynamicGlycopeptideMSMSAnalysisSerializer(GlycopeptideMSMSAnalysisSerializer):
    def _get_glycan_combination_for_glycopeptide(self, glycopeptide_id) -> int:
        return glycopeptide_id.glycan_combination_id

    def _get_peptide_id_for_glycopeptide(self, glycopeptide_id) -> int:
        return glycopeptide_id.peptide_id

    def fetch_glycopeptides(self, glycopeptide_ids) -> List[Glycopeptide]:
        aggregate = dict()
        for gp in self._identified_glycopeptide_set:
            for solution_set in gp.spectrum_matches:
                for match in solution_set:
                    aggregate[match.target.id] = match.target
        out = []
        for i, obj in enumerate(aggregate.values(), 1):
            inst = Glycopeptide(
                id=obj.id,
                peptide_id=obj.id.peptide_id,
                glycan_combination_id=obj.id.glycan_combination_id,
                protein_id=obj.id.protein_id,
                hypothesis_id=obj.id.hypothesis_id,
                glycopeptide_sequence=obj.get_sequence(),
                calculated_mass=obj.total_mass,
                formula=formula(obj.total_composition()))
            out.append(inst)
        return out

    def _migrate_single_glycopeptide(self, glycopeptide: Glycopeptide) -> int:
        inst = Glycopeptide(
            id=glycopeptide.id,
            peptide_id=glycopeptide.id.peptide_id,
            glycan_combination_id=glycopeptide.id.glycan_combination_id,
            protein_id=glycopeptide.id.protein_id,
            hypothesis_id=glycopeptide.id.hypothesis_id,
            glycopeptide_sequence=glycopeptide.get_sequence(),
            calculated_mass=glycopeptide.total_mass,
            formula=formula(glycopeptide.total_composition()))
        self._glycopeptide_hypothesis_migrator.migrate_glycopeptide(inst)
        self._glycopeptide_hypothesis_migrator.commit()
        return self._glycopeptide_hypothesis_migrator.glycopeptide_id_map[glycopeptide.id]



cimport cython
from cpython.list cimport PyList_Append, PyList_Size, PyList_GetItem

import numpy as np
cimport numpy as np

from ms_deisotope._c.peak_set cimport DeconvolutedPeak
from glypy.composition.ccomposition cimport CComposition
from glycresoft._c.chromatogram_tree.mass_shift cimport MassShiftBase, MassShift

from glypy.utils import uid

cdef MassShiftBase Unmodified = MassShift("Unmodified", CComposition())


cdef class ChromatogramTreeNode(object):
    def __init__(self, retention_time=None, scan_id=None, children=None, members=None,
                 node_type=Unmodified):
        if children is None:
            children = []
        if members is None:
            members = []
        self.retention_time = retention_time
        self.scan_id = scan_id
        self.children = children
        self.members = members
        self.node_type = node_type
        self._most_abundant_member = None
        self._neutral_mass = 0
        self._charge_states = set()
        self._recalculate()
        self._has_msms = None
        self.node_id = uid()

    def __reduce__(self):
        return self.__class__, (
            self.retention_time, self.scan_id, [c for c in self.children],
            list(self.members), self.node_type), self.__getstate__()

    def __getstate__(self):
        return {
            "node_id": self.node_id
        }

    def __setstate__(self, state):
        self.node_id = state['node_id']

    cpdef ChromatogramTreeNode clone(self):
        node = ChromatogramTreeNode(
            self.retention_time, self.scan_id, [c.clone() for c in self.children],
            list(self.members), node_type=self.node_type)
        node.node_id = self.node_id
        return node

    def _unspool_strip_children(self):
        node = ChromatogramTreeNode(
            self.retention_time, self.scan_id, [], list(self.members), node_type=self.node_type)
        yield node
        for child in self.children:
            for node in child._unspool_strip_children():
                yield node

    cpdef _recalculate(self):
        self._calculate_most_abundant_member()
        self._neutral_mass = self._most_abundant_member.neutral_mass
        self._charge_states = None
        self._has_msms = None

    cpdef _calculate_most_abundant_member(self):
        cdef:
            size_t i, n
            DeconvolutedPeak peak, candidate
        n = PyList_Size(self.members)
        if n == 1:
            self._most_abundant_member = self.members[0]
        else:
            if n == 0:
                self._most_abundant_member = None
            else:
                candidate = <DeconvolutedPeak>PyList_GetItem(self.members, 0)
                for i in range(1, n):
                    peak = <DeconvolutedPeak>PyList_GetItem(self.members, i)
                    if peak.intensity > candidate.intensity:
                        candidate = peak
                self._most_abundant_member = candidate

    @property
    def neutral_mass(self):
        if self._neutral_mass == 0:
            if self._most_abundant_member is not None:
                self._neutral_mass = self.get_neutral_mass()
        return self._neutral_mass

    cdef double get_neutral_mass(self):
        if self._most_abundant_member is not None:
            return self._most_abundant_member.neutral_mass
        else:
            return 0.0

    cpdef ChromatogramTreeNode _find(self, MassShiftBase node_type=Unmodified):
        if self.node_type == node_type:
            return self
        else:
            for child in self.children:
                match = child._find(node_type)
                if match is not None:
                    return match

    cpdef ChromatogramTreeNode find(self, MassShiftBase node_type=Unmodified):
        match = self._find(node_type)
        if match is not None:
            return match
        else:
            raise KeyError(node_type)

    cdef set get_contained_charge_states(self):
        cdef:
            set charges
            DeconvolutedPeak peak
            size_t i, n
        charges = set()
        n = PyList_Size(self.members)
        for i in range(n):
            peak = <DeconvolutedPeak>PyList_GetItem(self.members, i)
            charges.add(peak.charge)
        return charges

    cdef set get_charge_states(self):
        cdef:
            set charges
            size_t i, n
            ChromatogramTreeNode node

        charges = self.get_contained_charge_states()
        n = PyList_Size(self.children)
        for i in range(n):
            node = <ChromatogramTreeNode>PyList_GetItem(self.children, i)
            charges.update(node.get_charge_states())
        return charges

    @property
    def _contained_charge_states(self):
        if self._charge_states is None:
            self._charge_states = self.get_contained_charge_states()
        return self._charge_states

    cpdef set charge_states(self):
        return self.get_charge_states()

    @property
    def has_msms(self):
        if self._has_msms is None:
            self._has_msms = self._has_any_peaks_with_msms()
        return self._has_msms

    cdef double _total_intensity_members(self):
        total = 0.
        for peak in self.members:
            total += peak.intensity
        return total

    cdef double _total_intensity_children(self):
        total = 0.
        for child in self.children:
            total += child.total_intensity()
        return total

    cpdef double max_intensity(self):
        return self._most_abundant_member.intensity

    cpdef double total_intensity(self):
        return self._total_intensity_children() + self._total_intensity_members()

    def __eq__(self, other):
        return self.members == other.members

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.uid)

    def _has_any_peaks_with_msms(self):
        for peak in self.members:
            if peak.chosen_for_msms:
                return True
        for child in self.children:
            if child._has_any_peaks_with_msms():
                return True
        return False

    @property
    def peaks(self):
        peaks = list(self.members)
        for child in self.children:
            peaks.extend(child.peaks)
        return peaks


from uuid import uuid4

from glycresoft.serialize import (
    GlycanHypothesis, GlycanComposition,
    DatabaseBoundOperation, GlycanCompositionToClass)

from glycresoft.serialize.utils import get_uri_for_instance

from .glycan_source import GlycanClassLoader

from glycresoft.task import TaskBase


class GlycanHypothesisMigrator(DatabaseBoundOperation, TaskBase):
    def __init__(self, database_connection):
        DatabaseBoundOperation.__init__(self, database_connection)
        self.hypothesis_id = None
        self.glycan_composition_id_map = dict()
        self._structure_class_loader = None

    def make_structure_class_loader(self):
        return GlycanClassLoader(self.session)

    @property
    def structure_class_loader(self):
        if self._structure_class_loader is None:
            self._structure_class_loader = self.make_structure_class_loader()
        return self._structure_class_loader

    def migrate_hypothesis(self, hypothesis):
        new_hypothesis = GlycanHypothesis(
            name=hypothesis.name,
            uuid=str(uuid4().hex),
            status=hypothesis.status,
            parameters=dict(hypothesis.parameters or {}))
        new_hypothesis.parameters['original_uuid'] = hypothesis.uuid
        new_hypothesis.parameters['copied_from'] = get_uri_for_instance(hypothesis)
        self.session.add(new_hypothesis)
        self.session.flush()
        self.hypothesis_id = new_hypothesis.id

    def _migrate(self, obj):
        self.session.add(obj)
        self.session.flush()
        new_id = obj.id
        return new_id

    def commit(self):
        self.session.commit()

    def migrate_glycan_composition(self, glycan_composition):
        gc = glycan_composition
        new_gc = GlycanComposition(
            calculated_mass=gc.calculated_mass,
            formula=gc.formula,
            composition=gc.composition,
            hypothesis_id=self.hypothesis_id)
        new_id = self._migrate(new_gc)
        self.glycan_composition_id_map[gc.id] = new_id

        class_types = []
        for glycan_class in gc.structure_classes:
            local_glycan_class = self.structure_class_loader[glycan_class.name]
            class_types.append({"glycan_id": new_id, "class_id": local_glycan_class.id})
        if class_types:
            self.session.execute(GlycanCompositionToClass.insert(), class_types)



from collections import OrderedDict

from six import string_types as basestring

from matplotlib import patches as mpatches

from glypy.structure.glycan_composition import FrozenGlycanComposition, FrozenMonosaccharideResidue
from glycopeptidepy.utils import simple_repr

from glycresoft.database.composition_network import (
    CompositionRangeRule,
    CompositionRuleClassifier,
    CompositionRatioRule,
    normalize_composition)


def _degree_monosaccharide_alteration(x):
    try:
        if not isinstance(x, FrozenMonosaccharideResidue):
            x = FrozenMonosaccharideResidue.from_iupac_lite(str(x))
        return (len(x.modifications), len(x.substituent_links))
    except Exception:
        return (float('inf'), float('inf'))


class GlycanCompositionOrderer(object):
    def __init__(self, priority_residues=None, sorter=None):
        self.priority_residues = priority_residues or []
        self.sorter = sorter or _degree_monosaccharide_alteration

        self.priority_residues = [
            residue if isinstance(
                residue, FrozenMonosaccharideResidue) else FrozenMonosaccharideResidue.from_iupac_lite(
                residue)
            for residue in self.priority_residues
        ]

        outer = self

        class _ComparableProxy(object):
            def __init__(self, composition, obj=None):
                if isinstance(composition, basestring):
                    composition = FrozenGlycanComposition.parse(composition)
                if obj is None:
                    obj = composition
                self.composition = composition
                self.obj = obj

            def __iter__(self):
                return iter(self.composition)

            def __getitem__(self, key):
                return self.composition[key]

            def __lt__(self, other):
                return outer(self, other) < 0

            def __gt__(self, other):
                return outer(self, other) > 0

            def __eq__(self, other):
                return outer(self, other) == 0

            def __le__(self, other):
                return outer(self, other) <= 0

            def __ge__(self, other):
                return outer(self, other) >= 0

            def __ne__(self, other):
                return outer(self, other) != 0

        self._comparable_proxy = _ComparableProxy

    def sort(self, compositions, key=None, reverse=False):
        if key is None:

            def key(x):  # pylint: disable=function-redefined
                return x

        proxies = [self._comparable_proxy(key(c), c) for c in compositions]
        proxies = sorted(proxies, reverse=reverse)
        out = [p.obj for p in proxies]
        return out

    def key_order(self, keys):
        keys = list(keys)
        for r in reversed(self.priority_residues):
            try:
                i = keys.index(r)
                keys.pop(i)
                keys = [r] + keys
            except ValueError:
                pass
        return keys

    def __call__(self, a, b):
        if isinstance(a, basestring):
            a = normalize_composition(a)
            b = normalize_composition(b)
        keys = self.key_order(sorted(set(a) | set(b), key=self.sorter))

        for key in keys:
            if a[key] < b[key]:
                return -1
            elif a[key] > b[key]:
                return 1
            else:
                continue
        return 0

    __repr__ = simple_repr


class allset(frozenset):
    def __contains__(self, k):
        return True


class GlycanCompositionClassifierColorizer(object):
    def __init__(self, rule_color_map=None, default=None):
        self.rule_color_map = rule_color_map or {}
        self.default = default

    def __call__(self, obj):
        obj = normalize_composition(obj)
        for rule, color in self.rule_color_map.items():
            if rule(obj):
                return color
        if self.default:
            return self.default
        raise ValueError("Could not classify %r" % obj)

    def classify(self, obj):
        for rule, color in self.rule_color_map.items():
            if rule(obj):
                return rule.name
        return None

    __repr__ = simple_repr

    def make_legend(self, included=allset(), alpha=0.5):
        return [
            mpatches.Patch(
                label=rule.name, color=color, alpha=alpha) for rule, color in self.rule_color_map.items()
            if rule.name in included
        ]


NGlycanCompositionColorizer = GlycanCompositionClassifierColorizer(OrderedDict([
    (CompositionRuleClassifier("Sulfated Hybrid", [
     CompositionRangeRule("HexNAc", 3, 3) & CompositionRangeRule("@sulfate", 1)]), 'yellow'),
    (CompositionRuleClassifier("Sulfated Bi-Antennary",
                               [CompositionRangeRule("HexNAc", 4, 4) & CompositionRangeRule("@sulfate", 1)]), 'teal'),
    (CompositionRuleClassifier("Sulfated Tri-Antennary",
                               [CompositionRangeRule("HexNAc", 5, 5) & CompositionRangeRule("@sulfate", 1)]), 'rosybrown'),
    (CompositionRuleClassifier("Sulfated Tetra-Antennary",
                               [CompositionRangeRule("HexNAc", 6, 6) & CompositionRangeRule("@sulfate", 1)]), 'magenta'),
    (CompositionRuleClassifier("Paucimannose", [
        CompositionRangeRule("HexNAc", 2, 2) & CompositionRangeRule("Hex", 0, 4)]), "#f05af0"),
    (CompositionRuleClassifier("High Mannose", [
     CompositionRangeRule("HexNAc", 2, 2)]), '#1f77b4'),
    (CompositionRuleClassifier("Hybrid", [
     CompositionRangeRule("HexNAc", 3, 3)]), '#ff7f0e'),
    (CompositionRuleClassifier("Bi-Antennary",
                               [CompositionRangeRule("HexNAc", 4, 4)]), '#2ca02c'),
    (CompositionRuleClassifier("Tri-Antennary",
                               [CompositionRangeRule("HexNAc", 5, 5)]), '#d62728'),
    (CompositionRuleClassifier("Tetra-Antennary",
                               [CompositionRangeRule("HexNAc", 6, 6)]), '#9467bd'),
    (CompositionRuleClassifier("Penta-Antennary",
                               [CompositionRangeRule("HexNAc", 7, 7)]), '#8c564b'),
    (CompositionRuleClassifier("Supra-Penta-Antennary",
                               [CompositionRangeRule("HexNAc", 8)]), 'brown'),
    (CompositionRuleClassifier("Low Sulfate GAG", [
     CompositionRatioRule("HexN", "@sulfate", (0, 2))]), "#2aaaaa"),
    (CompositionRuleClassifier("High Sulfate GAG", [
     CompositionRatioRule("HexN", "@sulfate", (2, 4))]), "#88faaa")
]), default="slateblue")

NGlycanCompositionOrderer = GlycanCompositionOrderer(["HexNAc", "Hex", "Fuc", "NeuAc"])

_null_color_chooser = GlycanCompositionClassifierColorizer({}, default='blue')


class GlycanLabelTransformer(object):
    def __init__(self, label_series, order_chooser):
        self._input_series = label_series
        self.order_chooser = order_chooser
        self.residues = None
        self._infer_compositions()

    def _infer_compositions(self):
        residues = set()
        for item in self._input_series:
            if isinstance(item, basestring):
                item = normalize_composition(item)
            residues.update(item)

        residues = sorted(residues, key=_degree_monosaccharide_alteration)
        self.residues = self.order_chooser.key_order(residues)

    def transform(self):
        for item in self._input_series:
            if isinstance(item, basestring):
                item = normalize_composition(item)
            counts = [str(item[r]) for r in self.residues]
            yield '[%s]' % '; '.join(counts)

    __iter__ = transform

    __repr__ = simple_repr

    @property
    def label_key(self):
        return "Key: [%s]" % '; '.join(map(str, self.residues))



from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from scipy.ndimage import gaussian_filter1d
import numpy as np


from ..scoring import total_intensity


class LCMSSurfaceArtist(object):
    def __init__(self, chromatograms):
        self.chromatograms = chromatograms
        self.times = []
        self.masses = []
        self.heights = []

    def build_map(self):
        self.times = []
        self.masses = []
        self.heights = []

        for chroma in self.chromatograms:
            x, z = chroma.as_arrays()
            y = chroma.neutral_mass
            self.times.append(x)
            self.masses.append(y)

        rt = set()
        map(rt.update, self.times)
        rt = np.array(list(rt))
        rt.sort()
        self.times = rt

        self.heights = list(map(self.make_z_array, self.chromatograms))
        scaler = max(map(max, self.heights)) / 100.
        for height in self.heights:
            height /= scaler

    def make_z_array(self, chroma):
        z = []
        next_time_i = 0
        next_time = chroma.retention_times[next_time_i]

        for i in self.times:
            if np.allclose(i, next_time):
                z.append(total_intensity(chroma.peaks[next_time_i]))
                next_time_i += 1
                if next_time_i == len(chroma):
                    break
                next_time = chroma.retention_times[next_time_i]
            else:
                z.append(0)
        z = gaussian_filter1d(np.concatenate((z, np.zeros(len(self.times) - len(z)))), 1)
        return z

    def make_sparse(self, width=0.05):
        i = 0
        masses = []
        heights = []

        flat = self.heights[0] * 0

        masses.append(self.masses[0] - 200)
        heights.append(flat)

        while i < len(self.masses):
            mass = self.masses[i]
            masses.append(mass - width)
            heights.append(flat)
            masses.append(mass)
            heights.append(self.heights[i])
            masses.append(mass + width)
            heights.append(flat)
            i += 1

        self.masses = masses
        self.heights = heights

    def draw(self, alpha=0.8, **kwargs):
        fig = plt.figure()
        ax = fig.gca(projection='3d')
        self.ax = ax

        self.build_map()
        self.make_sparse()

        X, Y = np.meshgrid(self.times, self.masses)
        ax.plot_surface(X, Y, self.heights, rstride=1, cstride=1,
                        linewidth=0, antialiased=False, shade=True,
                        alpha=alpha)
        ax.view_init()
        ax.azim += 20
        ax.set_xlim3d(self.times.min(), self.times.max())
        ax.set_ylim3d(min(self.masses) - 100, max(self.masses))
        ax.set_xlabel("Retention Time (Min)", fontsize=18)
        ax.set_ylabel("Neutral Mass", fontsize=18)
        ax.set_zlabel("Relative Abundance (%)", fontsize=18)
        return self



import unittest
import tempfile

from io import StringIO

from glycresoft.database.builder.glycan import constrained_combinatorics


FILE_SOURCE = '''Hex 3 12
HexNAc 2 10
Fuc 0 5
Neu5Ac 0 4

Fuc < HexNAc
HexNAc > NeuAc + 1
'''      # The mismatch between Neu5Ac <-> NeuAc tests the IUPAC normalization mechanism


FILE_SOURCE2 = '''Hex 3 12
HexNAc 2 10
Fuc 0 5
Neu5Ac 0 4

Fuc < HexNAc
HexNAc - 1 > NeuAc
'''

FILE_SOURCE3 = '''Hex 3 12
HexNAc 2 10
Fuc 0 5
Neu5Ac 0 4

Fuc < HexNAc
NeuAc < HexNAc - 1
'''

FILE_SOURCE4 = '''Hex 3 12
HexNAc 2 10
Fuc 0 5
Neu5Ac 0 4

Fuc < HexNAc
NeuAc < HexNAc
'''


class GlycanCombinatoricsTests(unittest.TestCase):

    def setup_tempfile(self):
        file_name = tempfile.mktemp()
        file_name += '.tmp'
        open(file_name, 'w').write(FILE_SOURCE)
        return file_name

    def clear_file(self, path):
        open(path, 'wb').close()

    def test_run(self):
        file_name = self.setup_tempfile()
        builder = constrained_combinatorics.CombinatorialGlycanHypothesisSerializer(
            file_name, file_name + '.db')
        builder.run()
        inst = builder.query(constrained_combinatorics.DBGlycanComposition).filter(
            constrained_combinatorics.DBGlycanComposition.hypothesis_id == builder.hypothesis_id,
            constrained_combinatorics.DBGlycanComposition.composition == "{Hex:3; HexNAc:2}").one()
        self.assertAlmostEqual(inst.calculated_mass, 910.32777, 3)

        self.do_neuac_check(builder)

        builder.engine.dispose()
        self.clear_file(file_name + '.db')
        self.clear_file(file_name)

    def do_neuac_check(self, builder):
        i = 0
        for composition in builder.query(constrained_combinatorics.DBGlycanComposition):
            composition = composition.convert()
            i += 1
            self.assertGreater(composition["HexNAc"] - 1, composition['Neu5Ac'])


class GlycanCombinatoricsTestsWithReversedTermInConditional(GlycanCombinatoricsTests):

    def setup_tempfile(self):
        file_name = tempfile.mktemp()
        open(file_name, 'w').write(FILE_SOURCE2)
        return file_name


class GlycanCombinatoricsTestsWithReversedConditional(GlycanCombinatoricsTests):

    def setup_tempfile(self):
        file_name = tempfile.mktemp()
        open(file_name, 'w').write(FILE_SOURCE3)
        return file_name


class GlycanCombinatoricsTestsWithoutSubtractionInConditional(GlycanCombinatoricsTests):

    def setup_tempfile(self):
        file_name = tempfile.mktemp()
        open(file_name, 'w').write(FILE_SOURCE4)
        return file_name

    def do_neuac_check(self, builder):
        self.assertRaises(AssertionError, lambda: super(
            GlycanCombinatoricsTestsWithoutSubtractionInConditional, self).do_neuac_check(builder))


if __name__ == '__main__':
    unittest.main()



from ms_deisotope._c.peak_set cimport DeconvolutedPeak, DeconvolutedPeakSet

cdef class SignatureSpecification(object):
    cdef:
        public tuple components
        public tuple masses
        public bint _is_compound
        public Py_hash_t _hash
        double* _masses
        size_t n_masses

    cdef void _init_mass_array(self)
    cpdef bint is_expected(self, glycan_composition)
    cpdef int count_of(self, glycan_composition)
    cpdef DeconvolutedPeak peak_of(self, DeconvolutedPeakSet spectrum, double error_tolerance)


cdef class SignatureIonMatchRecord:
    cdef:
        public dict expected_matches
        public dict unexpected_matches
        public object glycan_composition
        public double score

    @staticmethod
    cdef SignatureIonMatchRecord _create(dict expected_matches, dict unexpected_matches, object glycan_composition)
    cpdef match(self, list signatures, DeconvolutedPeakSet spectrum, double error_tolerance=*)


cdef class SignatureIonIndexMatch:
    cdef:
        public dict glycan_to_key
        public dict key_to_record
        public double base_peak_intensity

    @staticmethod
    cdef SignatureIonIndexMatch _create(dict glycan_to_key, dict key_to_record, double base_peak_intensity)

    cpdef SignatureIonMatchRecord record_for(self, object glycan_composition)


cdef class SignatureIonIndex:
    cdef:
        public list signatures
        public dict glycan_to_key
        public dict key_to_representative

    cpdef SignatureIonIndexMatch match(self, DeconvolutedPeakSet spectrum, double error_tolerance=*)



cdef class OxoniumIndexMatch(object):
    cdef:
        public dict index_matches
        public dict glycan_to_index
        public dict id_to_index

    @staticmethod
    cdef OxoniumIndexMatch _create(dict index_matches, dict glycan_to_index, dict id_to_index)

    cpdef list by_glycan(self, glycan)
    cpdef list by_id(self, glycan_id)


cdef class OxoniumIndex(object):
    '''An index for quickly matching all oxonium ions against a spectrum and efficiently mapping them
    back to individual glycan compositions.
    '''
    cdef:
        public list fragments
        public dict fragment_index
        public dict glycan_to_index
        public dict index_to_glycan
        public dict index_to_simplified_index

    cpdef object simplify(self)
    cpdef OxoniumIndexMatch match(self, DeconvolutedPeakSet spectrum, double error_tolerance)


import io
import os
import re
import bisect

from collections import defaultdict, OrderedDict, namedtuple, deque
from typing import Any, Set, List, Union, Dict, Optional

from brainpy import mass_charge_ratio

from glypy.composition import formula
from glypy.io.nomenclature import identity
from glypy.structure.glycan_composition import (
    MonosaccharideResidue,
    FrozenMonosaccharideResidue, SubstituentResidue, FrozenGlycanComposition)

from glycopeptidepy.structure import parser, modification, PeptideSequence

from psims.mzid import components
from psims.mzid.writer import MzIdentMLWriter
from psims.controlled_vocabulary.controlled_vocabulary import load_gno

from ms_deisotope.output import mzml, ProcessedMSFileLoader

from glycresoft import task, serialize, version
from glycresoft.chromatogram_tree import Unmodified
from glycresoft.chromatogram_tree.chromatogram import group_by
from glycresoft.structure import FragmentCachingGlycopeptide


GlycopeptideType = Union[PeptideSequence, FragmentCachingGlycopeptide, serialize.Glycopeptide]
Props = Dict[str, Any]


class mass_term_pair(namedtuple("mass_term_pair", ('mass', 'term'))):
    def __lt__(self, other):
        return self.mass < float(other)

    def __gt__(self, other):
        return self.mass > float(other)

    def __float__(self):
        return self.mass


valid_monosaccharide_names = [
    "Hex",
    "HexNAc",
    "dHex",
    "NeuAc",
    "NeuGc",
    "Pen",
    "Fuc",
    # "HexA",
    # "HexN",
]

valid_monosaccharides = {FrozenMonosaccharideResidue.from_iupac_lite(v): v
                         for v in valid_monosaccharide_names}


def monosaccharide_to_term(monosaccharide):
    try:
        return valid_monosaccharides[monosaccharide]
    except KeyError:
        value = str(monosaccharide)
        return value


substituent_map = {
    "S": "sulfate",
    "P": "phosphate",
    # "Me": "methyl",
    # "Ac": "acetyl",
}

inverted_substituent_map = {
    v: k for k, v in substituent_map.items()
}

substituent_map['Sulpho'] = "sulfate"
substituent_map['Phospho'] = "phosphate"


def mparam(name, value=None, accession=None, cvRef="PSI-MS", **kwargs):
    if isinstance(name, dict):
        value = name.pop('value', None)
        accession = name.pop('accession')
        cvRef = name.pop('cvRef', cvRef)
        name_ = name.pop("name")
        kwargs.update(kwargs)
        name = name_
    return components.CVParam(
        name=name,
        value=value,
        accession=accession,
        cvRef=cvRef,
        **kwargs)


def parse_glycan_formula(glycan_formula):
    gc = FrozenGlycanComposition()
    if glycan_formula.startswith("\""):
        glycan_formula = glycan_formula[1:-1]
    for mono, count in re.findall(r"([^0-9]+)\((\d+)\)", glycan_formula):
        count = int(count)
        if mono in substituent_map:
            parsed = SubstituentResidue(substituent_map[mono])
        elif mono in ("Sia", ):
            continue
        elif mono in ("Pent", ):
            mono = "Pen"
            parsed = FrozenMonosaccharideResidue.from_iupac_lite(mono)
        elif mono == 'Xxx':
            continue
        elif mono == 'X':
            continue
        else:
            parsed = FrozenMonosaccharideResidue.from_iupac_lite(mono)
        gc[parsed] += count
    return gc


class GNOmeResolver(object):
    def __init__(self, cv=None):
        if cv is None:
            cv = load_gno()
        self.cv = cv
        self.build_mass_search_index()
        self.add_glycan_compositions()

    def add_glycan_compositions(self):
        formula_key = "GNO:00000202"
        for term in self.cv.terms.values():
            glycan_formula = term.get(formula_key)
            if glycan_formula:
                term['glycan_composition'] = parse_glycan_formula(glycan_formula)

    def build_mass_search_index(self):
        mass_index = []

        for term in self.cv.terms.values():
            match = re.search(r"weight of (\d+\.\d+) Da", term.definition)
            if match:
                mass = float(match.group(1))
                term['mass'] = mass
                mass_index.append(mass_term_pair(mass, term))

        mass_index.sort()
        self.mass_index = mass_index

    def _find_mass_match(self, mass):
        i = bisect.bisect_left(self.mass_index, mass)
        lo = self.mass_index[i - 1]
        lo_err = abs(lo.mass - mass)
        hi = self.mass_index[i]
        hi_err = abs(hi.mass - mass)
        if hi_err < lo_err:
            term = hi.term
        elif hi_err > lo_err:
            term = lo.term
        else:
            raise ValueError(
                "Ambiguous duplicate masses (%0.2f, %0.2f)" % (lo.mass, hi.mass))
        return term

    def resolve_gnome(self, glycan_composition):
        mass = glycan_composition.mass()
        term = self._find_mass_match(mass)
        recast = glycan_composition.clone().reinterpret(valid_monosaccharides)
        visit_queue = deque(term.children)
        while visit_queue:
            child = visit_queue.popleft()
            gc = child.get("glycan_composition")
            if gc is None:
                visit_queue.extend(child.children)
            elif gc == recast:
                return child

    def glycan_composition_to_terms(self, glycan_composition):
        out = []
        term = self.resolve_gnome(glycan_composition)
        if term is not None:
            out.append({
                "accession": term.id,
                "name": term.name,
                "cvRef": term.vocabulary.name
            })
        reinterpreted = glycan_composition.clone().reinterpret(valid_monosaccharides)
        for mono, count in reinterpreted.items():
            if isinstance(mono, SubstituentResidue):
                subst = inverted_substituent_map.get(
                    mono.name.replace("@", ""))
                if subst is not None:
                    out.append({
                        "name": "monosaccharide count",
                        "value": ("%s:%d" % (subst, count)),
                        "accession": "MS:XXXXX2",
                        "cvRef": "PSI-MS"
                    })
                else:
                    out.append({
                        "name": "unknown monosaccharide count",
                        "value": ("%s:%0.3f:%d" % (mono.name.replace("@", ""), mono.mass(), count)),
                        "accession": "MS:XXXXX3",
                        "cvRef": "PSI-MS"
                    })
            elif isinstance(mono, MonosaccharideResidue):
                for known in valid_monosaccharides:
                    if identity.is_a(mono, known):
                        out.append({
                            "name": "monosaccharide count",
                            "value": ("%s:%d" % (monosaccharide_to_term(known), count)),
                            "accession": "MS:XXXXX2",
                            "cvRef": "PSI-MS"
                        })
                        break
                else:
                    out.append({
                        "name": "unknown monosaccharide count",
                        "value": ("%s:%0.3f:%d" % (monosaccharide_to_term(mono), mono.mass(), count)),
                        "accession": "MS:XXXXX3",
                        "cvRef": "PSI-MS"
                    })
            else:
                raise TypeError("Cannot handle unexpected component of type %s" % (type(mono), ))
        return out


class SequenceIdTracker(object):
    mapping: Dict[str, int]

    def __init__(self):
        self.mapping = dict()

    def convert(self, glycopeptide: GlycopeptideType) -> int:
        s = str(glycopeptide)
        if s in self.mapping:
            return self.mapping[s]
        else:
            self.mapping[s] = glycopeptide.id
            return self.mapping[s]

    def __call__(self, glycopeptide: GlycopeptideType) -> int:
        return self.convert(glycopeptide)

    def dump(self):
        for key, value in self.mapping.items():
            print(value, key)


class MzMLExporter(task.TaskBase):
    def __init__(self, source, outfile):
        self.reader = ProcessedMSFileLoader(source)
        self.outfile = outfile
        self.writer = None
        self.n_spectra = None

    def make_writer(self):
        self.writer = mzml.MzMLScanSerializer(
            self.outfile, sample_name=self.reader.sample_run.name,
            n_spectra=self.n_spectra)

    def aggregate_scan_bunches(self, scan_ids):
        scans = defaultdict(list)
        for scan_id in scan_ids:
            scan = self.reader.get_scan_by_id(scan_id)
            scans[scan.precursor_information.precursor_scan_id].append(
                scan)
        bunches = []
        for precursor_id, products in scans.items():
            products.sort(key=lambda x: x.scan_time)
            precursor = self.reader.get_scan_by_id(precursor_id)
            bunches.append(mzml.ScanBunch(precursor, products))
        bunches.sort(key=lambda bunch: bunch.precursor.scan_time)
        return bunches

    def begin(self, scan_bunches):
        self.n_spectra = sum(len(b.products) for b in scan_bunches) + len(scan_bunches)
        self.make_writer()
        for bunch in scan_bunches:
            self.put_scan_bunch(bunch)

    def put_scan_bunch(self, bunch):
        self.writer.save_scan_bunch(bunch)

    def extract_chromatograms_from_identified_glycopeptides(self, glycopeptide_list):
        by_chromatogram = group_by(
            glycopeptide_list, lambda x: (
                x.chromatogram.chromatogram if x.chromatogram is not None else None))
        i = 0
        for chromatogram, members in by_chromatogram.items():
            if chromatogram is None:
                continue
            self.enqueue_chromatogram(chromatogram, i, params=[
                {"name": "GlycReSoft:profile score", "value": members[0].ms1_score},
                {"name": "GlycReSoft:assigned entity", "value": str(members[0].structure)}
            ])
            i += 1

    def enqueue_chromatogram(self, chromatogram, chromatogram_id, params=None):
        if params is None:
            params = []
        chromatogram_data = dict()
        rt, signal = chromatogram.as_arrays()
        chromatogram_dict = OrderedDict(zip(rt, signal))
        chromatogram_data['chromatogram'] = chromatogram_dict
        chromatogram_data['chromatogram_type'] = 'selected ion current chromatogram'
        chromatogram_data['id'] = chromatogram_id
        chromatogram_data['params'] = params

        self.writer.chromatogram_queue.append(chromatogram_data)

    def complete(self):
        self.writer.complete()
        self.writer.format()


glycosylation_type_to_term_map = {
    "N-Linked": {
        "name": "N-glycan",
        "accession": "MS:XXXXX5",
        "cvRef": "PSI-MS",
    },
    "O-Linked": {
        "name": "mucin O-glycan",
        "accession": "MS:XXXXX6",
        "cvRef": "PSI-MS",
    },
    "GAG linker": {
        "name": "glycosaminoglycan",
        "accession": "MS:XXXXX7",
        "cvRef": "PSI-MS",
    },
}

glycosylation_type_to_term_map["N-Glycan"] = glycosylation_type_to_term_map["N-Linked"]
glycosylation_type_to_term_map["O-Glycan"] = glycosylation_type_to_term_map["O-Linked"]


def glycosylation_type_to_term(glycosylation_type):
    return glycosylation_type_to_term_map[glycosylation_type]


class MzIdentMLSerializer(task.TaskBase):
    outfile: Union[os.PathLike, io.IOBase]

    analysis: serialize.Analysis
    database_handle: serialize.DatabaseBoundOperation
    gnome_resolver: GNOmeResolver

    _id_tracker: SequenceIdTracker
    _glycopeptide_list: List[serialize.IdentifiedGlycopeptide]
    _peptide_evidence: List[Props]

    protein_list: List[serialize.Protein]
    glycan_list: List[serialize.GlycanCombination]

    scan_ids: Set[str]

    q_value_threshold: float
    ms2_score_threshold: float

    export_mzml: bool
    source_mzml_path: Optional[os.PathLike]
    output_mzml_path: Optional[os.PathLike]

    embed_protein_sequences: bool
    report_top_match_per_glycopeptide: bool

    def __init__(self, outfile: Union[os.PathLike, io.IOBase],
                 glycopeptide_list: List[serialize.IdentifiedGlycopeptide],
                 analysis: serialize.Analysis,
                 database_handle: serialize.DatabaseBoundOperation,
                 q_value_threshold: float=0.05,
                 ms2_score_threshold: float=0,
                 export_mzml: bool=True,
                 source_mzml_path: Optional[os.PathLike]=None,
                 output_mzml_path: Optional[os.PathLike]=None,
                 embed_protein_sequences: bool=True,
                 report_top_match_per_glycopeptide: bool=True):

        self.outfile = outfile

        self.database_handle = database_handle
        self.analysis = analysis
        self.gnome_resolver = GNOmeResolver()

        self._glycopeptide_list = glycopeptide_list
        self.protein_list = []
        self.glycan_list = []
        self._peptide_evidence = []
        self.scan_ids = set()
        self._id_tracker = SequenceIdTracker()

        self.q_value_threshold = q_value_threshold
        self.ms2_score_threshold = ms2_score_threshold

        self.export_mzml = export_mzml
        self.source_mzml_path = source_mzml_path
        self.output_mzml_path = output_mzml_path

        self.report_top_match_per_glycopeptide = report_top_match_per_glycopeptide
        self.embed_protein_sequences = embed_protein_sequences

    def _coerce_orm(self, obj):
        if isinstance(obj, serialize.Base):
            obj = obj.convert()
        return obj

    @property
    def glycopeptide_list(self):
        return self._glycopeptide_list

    def extract_proteins(self):
        self.protein_list = self.database_handle.query(serialize.Protein).all()

    def extract_glycans(self):
        self.glycan_list = self.database_handle.query(serialize.GlycanCombination).all()

    def convert_to_peptide_dict(self, glycopeptide: GlycopeptideType, id_tracker: SequenceIdTracker) -> Props:
        data = {
            "id": glycopeptide.id,
            "peptide_sequence": parser.strip_modifications(glycopeptide),
            "modifications": []
        }

        glycopeptide = self._coerce_orm(glycopeptide)

        i = 0
        # TODO: handle N-terminal and C-terminal modifications
        glycosylation_event_count = len(glycopeptide.glycosylation_manager)
        glycosylation_events_handled = 0
        for _pos, mods in glycopeptide:
            i += 1
            if not mods:
                continue
            else:
                mod = mods[0]
            if mod.rule.is_a("glycosylation"):
                glycosylation_events_handled += 1
                is_aggregate_stub = False
                mod_params = [
                    glycosylation_type_to_term(
                        str(mod.rule.glycosylation_type))
                ]
                if mod.rule.is_core:

                    mod_params.extend(
                        self.gnome_resolver.glycan_composition_to_terms(glycopeptide.glycan_composition.clone()))

                    mass = glycopeptide.glycan_composition.mass()
                    if glycosylation_event_count == 1:
                        mod_params.append({
                            "name": "glycan composition",
                            "cvRef": "PSI-MS",
                            "accession": "MS:XXXX14"
                        })
                    else:
                        mod_params.append({
                            "name": "glycan aggregate",
                            "cvRef": "PSI-MS",
                            "accession": "MS:XXXX15"
                        })
                        if glycosylation_events_handled > 1:
                            mass = 0
                            is_aggregate_stub = True

                    if not is_aggregate_stub:
                        mod_params.append({
                            "accession": 'MS:1000864',
                            "cvRef": "PSI-MS",
                            "name": "chemical formula",
                            "value": formula(glycopeptide.glycan_composition.total_composition()),
                        })

                else:
                    mod_params.append({
                        "accession": 'MS:1000864',
                        "cvRef": "PSI-MS",
                        "name": "chemical formula",
                        "value": formula(mod.rule.composition),
                    })
                    if mod.rule.is_composition:
                        mod_params.extend(self.gnome_resolver.glycan_composition_to_terms(mod.rule.glycan.clone()))
                        mod_params.append({
                            "name": "glycan composition",
                            "cvRef": "PSI-MS",
                            "accession": "MS:XXXX14"
                        })
                    else:
                        mod_params.append({
                            "name": "glycan structure",
                            "cvRef": "PSI-MS",
                            "accession": "MS:XXXXXXX"
                        })
                    mass = mod.mass

                mod_dict = {
                    "monoisotopic_mass_delta": mass,
                    "location": i,
                    # "name": "unknown modification",
                    "name": "glycosylation modification",
                    "params": [components.CVParam(**x) for x in mod_params]
                }
                data['modifications'].append(mod_dict)
            else:
                mod_dict = {
                    "monoisotopic_mass_delta": mod.mass,
                    "location": i,
                    "name": mod.name,
                }
                data['modifications'].append(mod_dict)
        return data

    def _encode_score_set(self, spectrum_match: serialize.GlycopeptideSpectrumMatch) -> List[Union[Props, components.CVParam]]:
        score_params = [
            mparam("GlycReSoft:peptide score",
                   spectrum_match.score_set.peptide_score, "MS:XXX10C"),
            mparam("GlycReSoft:glycan score",
                   spectrum_match.score_set.glycan_score, "MS:XXX10B"),
            mparam("GlycReSoft:glycan coverage",
                   spectrum_match.score_set.glycan_coverage, "MS:XXX10H"),
            mparam("GlycReSoft:joint q-value",
                   spectrum_match.q_value, "MS:XXX10G"),
            mparam("GlycReSoft:peptide q-value",
                   spectrum_match.q_value_set.peptide_q_value,
                   "MS:XXX10E"),
            mparam("GlycReSoft:glycan q-value",
                   spectrum_match.q_value_set.glycan_q_value, "MS:XXX10F"),
            mparam("GlycReSoft:glycopeptide q-value",
                   spectrum_match.q_value_set.glycopeptide_q_value, "MS:XXX10D"),
        ]
        return score_params

    def convert_to_identification_item_dict(self, spectrum_match: serialize.GlycopeptideSpectrumMatch,
                                            seen_targets: Optional[Set[int]] = None,
                                            id_tracker: Optional[SequenceIdTracker] = None) -> Props:
        if seen_targets is None:
            seen_targets = set()
        if spectrum_match.target.id not in seen_targets:
            return None
        charge = spectrum_match.scan.precursor_information.charge
        data = {
            "charge_state": charge,
            "experimental_mass_to_charge": mass_charge_ratio(
                spectrum_match.scan.precursor_information.neutral_mass, charge),
            "calculated_mass_to_charge": mass_charge_ratio(
                spectrum_match.target.total_mass, charge),
            "peptide_id": id_tracker(spectrum_match.target),
            "peptide_evidence_id": spectrum_match.target.id,
            "score": mparam({
                "name": "GlycReSoft:total score",
                "value": spectrum_match.score,
                "accession": "MS:XXX10A",
            }),
            "params": [
                components.CVParam(**{
                    "name": "glycan dissociating, peptide preserving",
                    "accession": "MS:XXX111", "cvRef": "PSI-MS"}),
                components.CVParam(**{
                    "name": "glycan eliminated, peptide dissociating",
                    "accession": "MS:XXX114", "cvRef": "PSI-MS"}),
                {
                    "name": "scan start time",
                    "value": spectrum_match.scan.scan_time,
                    "unit_name": "minute"
                }
            ],
            "id": spectrum_match.id
        }
        if spectrum_match.is_multiscore():
            score_params = self._encode_score_set(spectrum_match)
            data['params'].extend(score_params)
        else:
            data['params'].extend([
                mparam("GlycReSoft:glycopeptide q-value",
                    spectrum_match.q_value, "MS:XXX10D"),
            ])
        if spectrum_match.mass_shift.name != Unmodified.name:
            data['params'].append(
                mparam("GlycReSoft:mass shift", "%s:%0.3f:%0.3f" % (
                    spectrum_match.mass_shift.name,
                    spectrum_match.mass_shift.mass,
                    spectrum_match.mass_shift.tandem_mass),
                    "MS:XXX10I"))
        return data

    def convert_to_spectrum_identification_dict(self, spectrum_solution_set: serialize.GlycopeptideSpectrumSolutionSet,
                                                seen_targets: Optional[Set[int]] = None,
                                                id_tracker: Optional[SequenceIdTracker] = None) -> Props:
        data = {
            "spectra_data_id": 1,
            "spectrum_id": spectrum_solution_set.scan.scan_id,
            "id": spectrum_solution_set.id
        }
        idents = []
        for item in spectrum_solution_set:
            d = self.convert_to_identification_item_dict(
                item, seen_targets=seen_targets, id_tracker=id_tracker)
            if d is None:
                continue
            idents.append(d)
        data['identifications'] = idents
        return data

    def convert_to_peptide_evidence_dict(self, glycopeptide: GlycopeptideType, id_tracker: SequenceIdTracker) -> Props:
        data = {
            "start_position": glycopeptide.protein_relation.start_position,
            "end_position": glycopeptide.protein_relation.end_position,
            "peptide_id": id_tracker(glycopeptide),
            "db_sequence_id": glycopeptide.protein_relation.protein_id,
            "is_decoy": False,
            "id": glycopeptide.id
        }
        return data

    def convert_to_protein_dict(self, protein: serialize.Protein, include_sequence: bool = True) -> Props:
        data = {
            "id": protein.id,
            "accession": protein.name,
            "search_database_id": 1,
        }
        if include_sequence:
            data["sequence"] = protein.protein_sequence
        return data

    def extract_peptides(self):
        self.log("Extracting Proteins")
        self.extract_proteins()
        self._peptides = []
        seen = set()

        self.log("Extracting Peptides")
        for gp in self.glycopeptide_list:
            d = self.convert_to_peptide_dict(gp.structure, self._id_tracker)

            if self._id_tracker(gp.structure) == gp.structure.id:
                self._peptides.append(d)
                seen.add(gp.structure.id)

        self.log("Extracting PeptideEvidence")
        self._peptide_evidence = [
            self.convert_to_peptide_evidence_dict(
                gp.structure, self._id_tracker) for gp in self.glycopeptide_list
        ]

        self._proteins = [
            self.convert_to_protein_dict(prot, self.embed_protein_sequences)
            for prot in self.protein_list
        ]

    def extract_spectrum_identifications(self):
        self.log("Extracting SpectrumIdentificationResults")
        spectrum_identifications = []
        seen_scans = set()
        accepted_solution_ids = {gp.structure.id for gp in self.glycopeptide_list}
        gp_list = self.glycopeptide_list
        n = len(gp_list)
        j = 0
        for i, gp in enumerate(gp_list):
            if i % 25 == 0:
                self.log("... Processed %d glycopeptide features with %d spectra matched (%0.2f%%)" % (i, j, i * 100.0 / n))
            if self.report_top_match_per_glycopeptide:
                spectrum_matches: List[serialize.GlycopeptideSpectrumSolutionSet] = [
                    gp.best_spectrum_match.solution_set]
            else:
                spectrum_matches: List[serialize.GlycopeptideSpectrumSolutionSet] = gp.spectrum_matches
            for solution in spectrum_matches:
                j += 1
                if solution.scan.scan_id in seen_scans:
                    continue
                if solution.best_solution().q_value > self.q_value_threshold:
                    continue
                if solution.score < self.ms2_score_threshold:
                    continue
                seen_scans.add(solution.scan.scan_id)
                d = self.convert_to_spectrum_identification_dict(
                    solution, seen_targets=accepted_solution_ids,
                    id_tracker=self._id_tracker)
                if len(d['identifications']):
                    spectrum_identifications.append(d)

        self.scan_ids = seen_scans
        self._spectrum_identification_list = {
            "id": 1,
            "identification_results": spectrum_identifications
        }

    def software_entry(self) -> List[Props]:
        software = {
            "name": "GlycReSoft",
            "version": version.version,
            "uri": None
        }
        return [software]

    def search_database(self) -> Props:
        hypothesis = self.analysis.hypothesis
        spec = {
            "name": hypothesis.name,
            "location": self.database_handle._original_connection,
            "id": 1
        }
        if "fasta_file" in hypothesis.parameters:
            spec['file_format'] = 'fasta format'
            spec['location'] = hypothesis.parameters['fasta_file']
        elif "mzid_file" in hypothesis.parameters:
            spec['file_format'] = 'mzIdentML format'
        return spec

    def source_file(self) -> Props:
        spec = {
            "location": self.database_handle._original_connection,
            "file_format": "data stored in database",
            "id": 1
        }
        return spec

    def spectra_data(self) -> Props:
        spec = {
            "location": self.analysis.parameters['sample_path'],
            "file_format": 'mzML format',
            "spectrum_id_format": "multiple peak list nativeID format",
            "id": 1
        }
        return spec

    def protocol(self) -> Props:
        hypothesis = self.analysis.hypothesis
        analysis = self.analysis
        mods = []

        def transform_modification(mod):
            if isinstance(mod, str):
                mod_inst = modification.Modification(mod)
                target = modification.extract_targets_from_rule_string(mod)
                new_rule = mod_inst.rule.clone({target})
                return new_rule
            return mod

        def pack_modification(mod, fixed=True):
            mod_spec = {
                "fixed": fixed,
                "mass_delta": mod.mass,
                "residues": [res.symbol for rule in mod.targets
                             for res in rule.amino_acid_targets],
                "params": [
                    mod.name
                ]
            }
            return mod_spec

        def pack_glycan_modification(glycan_composition, fixed=False):
            params = self.gnome_resolver.glycan_composition_to_terms(self._coerce_orm(glycan_composition))
            glycan_types = sorted({str(b) for a in glycan_composition.component_classes for b in a})
            residues = []
            for gt in glycan_types:
                gt = str(gt)
                if gt == "N-Glycan":
                    residues.append("N")
                if gt == "O-Glycan":
                    residues.extend("ST")
                if gt == 'GAG-Linker':
                    residues.append("S")
                params.append(glycosylation_type_to_term(gt))
            if glycan_composition.count == 1:
                params.append({
                    "name": "glycan composition",
                    "cvRef": "PSI-MS",
                    "accession": "MS:XXXX14"
                })
            else:
                params.append({
                    "name": "glycan aggregate",
                    "cvRef": "PSI-MS",
                    "accession": "MS:XXXX15"
                })
            mod_spec = {
                "fixed": fixed,
                "mass_delta": glycan_composition.dehydrated_mass(),
                "name": "glycosylation modification",
                # "accession": "MS:XXXXX1",
                "residues": residues,
                "params": [components.CVParam(**x) for x in params]
            }
            return mod_spec


        for mod in hypothesis.parameters.get('constant_modifications', []):
            mod = transform_modification(mod)
            mods.append(pack_modification(mod, True))
        for mod in hypothesis.parameters.get('variable_modifications', []):
            mod = transform_modification(mod)
            mods.append(pack_modification(mod, False))

        for gc in sorted(self.glycan_list, key=lambda x: (x.mass(), x.id)):
            mods.append(pack_glycan_modification(gc, False))

        strategy = analysis.parameters.get("search_strategy")
        if strategy == "multipart-target-decoy-competition":
            fdr_params = [
                {"name": "glycopeptide false discovery rate control strategy",
                 "accession": "MS:XXX108", "cvRef": "PSI-MS"},
                {"name": "peptide glycopeptide false discovery rate control strategy",
                 "accession": "MS:XXX106", "cvRef": "PSI-MS"},
                {"name": "glycan glycopeptide false discovery rate control strategy",
                 "accession": "MS:XXX107", "cvRef": "PSI-MS"},
                {"name": "joint glycopeptide false discovery rate control strategy",
                 "accession": "MS:XXX11A", "cvRef": "PSI-MS"},
            ]
        else:
            fdr_params = [
                {"name": "glycopeptide false discovery rate control strategy",
                 "accession": "MS:XXX108", "cvRef": "PSI-MS"},
            ]
        spec = {
            "enzymes": [
                {"name": getattr(e, 'name', e), "missed_cleavages": hypothesis.parameters.get(
                    'max_missed_cleavages', None), "id": i}
                for i, e in enumerate(hypothesis.parameters.get('enzymes'))
            ],
            "fragment_tolerance": (analysis.parameters['fragment_error_tolerance'] * 1e6, None, "parts per million"),
            "parent_tolerance": (analysis.parameters['mass_error_tolerance'] * 1e6, None, "parts per million"),
            "modification_params": mods,
            "id": 1,
            "additional_search_params": [
                {
                    "name": "glycopeptide search",
                    "accession": "MS:XXXX98",
                    "cvRef": "PSI-MS",
                }
            ] + fdr_params + [
                {
                    "name": "param: b ion",
                    "accession": "MS:1001118",
                    "cvRef": "PSI-MS",
                },
                {
                    "name": "param: y ion",
                    "accession": "MS:1001262",
                    "cvRef": "PSI-MS",
                },
                {
                    "name": "param: peptide + glycan Y ion",
                    "accession": "MS:XXXX17",
                    "cvRef": "PSI-MS",
                },
                {
                    "name": "param: oxonium ion",
                    "accession": "MS:XXXX22",
                    "cvRef": "PSI-MS",
                },
            ]
        }
        spec['additional_search_params'] = [components.CVParam(**x) for x in spec['additional_search_params']]
        return spec

    def run(self):
        f = MzIdentMLWriter(self.outfile, vocabularies=[
            components.CV(
                id='GNO', uri="http://purl.obolibrary.org/obo/gno.obo", full_name='GNO'),
        ])
        self.log("Loading Spectra Data")
        spectra_data = self.spectra_data()
        self.log("Loading Search Database")
        search_database = self.search_database()
        self.log("Building Protocol")
        self.extract_glycans()
        protocol = self.protocol()
        source_file = self.source_file()
        self.extract_peptides()
        self.extract_spectrum_identifications()

        had_specified_mzml_path = self.source_mzml_path is None
        if self.source_mzml_path is None:
            self.source_mzml_path = spectra_data['location']

        if self.source_mzml_path is None:
            did_resolve_mzml_path = False
        else:
            did_resolve_mzml_path = os.path.exists(self.source_mzml_path)
        if not did_resolve_mzml_path:
            self.log("Could not locate source mzML file.")
            if not had_specified_mzml_path:
                self.log("If you did not specify an alternative location to "
                         "find the mzML path, please do so.")

        if self.export_mzml and did_resolve_mzml_path:
            if self.output_mzml_path is None:
                prefix = os.path.splitext(self.outfile.name)[0]
                self.output_mzml_path = "%s.export.mzML" % (prefix,)
            exporter = None
            self.log("Begin Exporting mzML")
            with open(self.output_mzml_path, 'wb') as handle:
                exporter = MzMLExporter(self.source_mzml_path, handle)
                self.log("... Aggregating Scan Bunches")
                scan_bunches = exporter.aggregate_scan_bunches(self.scan_ids)
                self.log("... Exporting Spectra")
                exporter.begin(scan_bunches)
                self.log("... Exporting Chromatograms")
                exporter.extract_chromatograms_from_identified_glycopeptides(
                    self.glycopeptide_list)
                self.log("... Finalizing mzML")
                exporter.complete()
            self.log("mzML Export Finished")

        analysis = [[spectra_data['id']], [search_database['id']]]

        with f:
            f.controlled_vocabularies()
            f.providence(software=self.software_entry())

            f.register("SpectraData", spectra_data['id'])
            f.register("SearchDatabase", search_database['id'])
            f.register("SpectrumIdentificationList", self._spectrum_identification_list['id'])

            with f.sequence_collection():
                for prot in self._proteins:
                    f.write_db_sequence(**prot)
                for pep in self._peptides:
                    f.write_peptide(**pep)
                for pe in self._peptide_evidence:
                    f.write_peptide_evidence(**pe)

            with f.analysis_protocol_collection():
                f.spectrum_identification_protocol(**protocol)

            with f.element("AnalysisCollection"):
                f.SpectrumIdentification(*analysis).write(f)

            with f.element("DataCollection"):
                f.inputs(source_file, search_database, spectra_data)
                with f.element("AnalysisData"):
                    with f.spectrum_identification_list(id=self._spectrum_identification_list['id']):
                        for result_ in self._spectrum_identification_list['identification_results']:
                            result = dict(result_)
                            identifications = result.pop("identifications")
                            result = f.spectrum_identification_result(**result)
                            with result:
                                for item in identifications:
                                    f.write_spectrum_identification_item(**item)

        f.outfile.close()



import importlib
import sys
from types import ModuleType

import glycresoft

class LazyModule(ModuleType):
    def __init__(self, name, mod_name):
        super().__init__(name)
        self.__mod_name = name

    def __getattr__(self, attr):
        if "_lazy_module" not in self.__dict__:
            self._lazy_module = importlib.import_module(self.__mod_name, package="glycresoft")
        return getattr(self._lazy_module, attr)

sys.modules['glycan_profiling'] = LazyModule("glycresoft", "glycresoft")
sys.modules["glycan_profiling.serialize"] = LazyModule("glycresoft.serialize", "glycresoft")
sys.modules["glycan_profiling.chromatogram_tree"] = LazyModule("glycresoft.chromatogram_tree", "glycresoft")
sys.modules["glycan_profiling.config"] = LazyModule("glycresoft.config", "glycresoft")
sys.modules["glycan_profiling.task"] = LazyModule("glycresoft.task", "glycresoft")
sys.modules["glycan_profiling.structure"] = LazyModule("glycresoft.structure", "glycresoft")




import math
from typing import List, Mapping, Optional, Set, Tuple, Type, Union


from glycopeptidepy import PeptideSequence
from glycopeptidepy.structure.fragment import ChemicalShift, IonSeries, FragmentBase
from glycopeptidepy.structure.fragmentation_strategy import EXDFragmentationStrategy, HCDFragmentationStrategy, PeptideFragmentationStrategyBase

from glycresoft.structure.fragment_match_map import FragmentMatchMap
from glycresoft.structure.structure_loader import FragmentCachingGlycopeptide
from ..core_search import glycan_side_group_count
from ...spectrum_match import SpectrumMatcherBase, ModelTreeNode


class GlycopeptideSpectrumMatcherBase(SpectrumMatcherBase):
    target: Union[FragmentCachingGlycopeptide, PeptideSequence]

    _peptide_Y_ion_utilization: Optional[Tuple[float, int, int]] = None
    _glycan_score: Optional[float] = None
    _peptide_score: Optional[float] = None
    _glycan_coverage: Optional[float] = None
    _peptide_coverage: Optional[float] = None

    solution_map: FragmentMatchMap

    def _theoretical_mass(self) -> float:
        return self.target.total_mass

    def _match_oxonium_ions(self, error_tolerance: float=2e-5, masked_peaks: Optional[Set[int]]=None) -> Set[int]:
        '''Note: This method is masked by a Cython implementation
        '''
        if masked_peaks is None:
            masked_peaks = set()
        for frag in self.target.glycan_fragments(
                all_series=False, allow_ambiguous=False,
                include_large_glycan_fragments=False,
                maximum_fragment_size=4):
            peak = self.spectrum.has_peak(frag.mass, error_tolerance)
            if peak and peak.index.neutral_mass not in masked_peaks:
                self.solution_map.add(peak, frag)
                masked_peaks.add(peak.index.neutral_mass)
        return masked_peaks

    def _match_stub_glycopeptides(self, error_tolerance: float=2e-5,
                                  masked_peaks: Optional[Set[int]]=None,
                                  chemical_shift: Optional[ChemicalShift]=None,
                                  extended_glycan_search: bool=False) -> Set[int]:
        if masked_peaks is None:
            masked_peaks = set()
        if not extended_glycan_search:
            fragments = self.target.stub_fragments(extended=True)
        else:
            fragments = self.target.stub_fragments(extended=True, extended_fucosylation=True)
        for frag in fragments:
            for peak in self.spectrum.all_peaks_for(frag.mass, error_tolerance):
                # should we be masking these? peptides which have amino acids which are
                # approximately the same mass as a monosaccharide unit at ther terminus
                # can produce cases where a stub ion and a backbone fragment match the
                # same peak.
                #
                masked_peaks.add(peak.index.neutral_mass)
                self.solution_map.add(peak, frag)
            if chemical_shift is not None:
                shifted_mass = frag.mass + self.mass_shift.tandem_mass
                for peak in self.spectrum.all_peaks_for(shifted_mass, error_tolerance):
                    masked_peaks.add(peak.index.neutral_mass)
                    shifted_frag = frag.clone()
                    shifted_frag.chemical_shift = chemical_shift
                    shifted_frag.name += "+ %s" % (self.mass_shift.name,)
                    self.solution_map.add(peak, shifted_frag)
        return masked_peaks

    def get_fragments(self, series: IonSeries,
                      strategy: Optional[Type[PeptideFragmentationStrategyBase]]=None,
                      **kwargs) -> List[List[FragmentBase]]:
        fragments = self.target.get_fragments(series, strategy=strategy)
        return fragments

    def _match_backbone_series(self, series: IonSeries, error_tolerance: float=2e-5,
                               masked_peaks: Optional[Set[int]]=None,
                               strategy: Optional[Type[PeptideFragmentationStrategyBase]] = None,
                               include_neutral_losses: bool=False):
        if strategy is None:
            strategy = HCDFragmentationStrategy
        if masked_peaks is None:
            masked_peaks = set()
        for frags in self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses):
            for frag in frags:
                for peak in self.spectrum.all_peaks_for(frag.mass, error_tolerance):
                    if peak.index.neutral_mass in masked_peaks:
                        continue
                    self.solution_map.add(peak, frag)

    def match(self, error_tolerance: float=2e-5, *args, **kwargs):
        masked_peaks = set()
        include_neutral_losses = kwargs.get("include_neutral_losses", False)
        extended_glycan_search = kwargs.get("extended_glycan_search", False)
        if self.mass_shift.tandem_mass != 0:
            chemical_shift = ChemicalShift(
                self.mass_shift.name, self.mass_shift.tandem_composition)
        else:
            chemical_shift = None

        is_hcd = self.is_hcd()
        is_exd = self.is_exd()
        if not is_hcd and not is_exd:
            is_hcd = True
        # handle glycan fragments from collisional dissociation
        if is_hcd:
            self._match_oxonium_ions(error_tolerance, masked_peaks=masked_peaks)
            self._match_stub_glycopeptides(error_tolerance, masked_peaks=masked_peaks,
                                           chemical_shift=chemical_shift,
                                           extended_glycan_search=extended_glycan_search)

        # handle N-term
        if is_hcd and not is_exd:
            self._match_backbone_series(
                IonSeries.b, error_tolerance, masked_peaks, HCDFragmentationStrategy, include_neutral_losses)
        elif is_exd:
            self._match_backbone_series(
                IonSeries.b, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)
            self._match_backbone_series(
                IonSeries.c, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)

        # handle C-term
        if is_hcd and not is_exd:
            self._match_backbone_series(
                IonSeries.y, error_tolerance, masked_peaks, HCDFragmentationStrategy,
                include_neutral_losses)
        elif is_exd:
            self._match_backbone_series(
                IonSeries.y, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)
            self._match_backbone_series(
                IonSeries.z, error_tolerance, masked_peaks, EXDFragmentationStrategy,
                include_neutral_losses)

        return self

    def peptide_score(self, *args, **kwargs) -> float:
        if self._peptide_score is None:
            self._peptide_score = self.calculate_peptide_score(*args, **kwargs)
        return self._peptide_score

    def calculate_peptide_score(self, *args, **kwargs) -> float:
        return 0

    def glycan_score(self, *args, **kwargs) -> float:
        if self._glycan_score is None:
            self._glycan_score = self.calculate_glycan_score(*args, **kwargs)
        return self._glycan_score

    def _glycan_side_group_count(self, glycan_composition: Mapping) -> int:
        return glycan_side_group_count(glycan_composition)

    def _calculate_glycan_coverage(self, core_weight: float=0.4,
                                   coverage_weight: float=0.5,
                                   fragile_fucose: bool=False,
                                   extended_glycan_search: bool=False) -> float:
        seen = set()
        series = IonSeries.stub_glycopeptide
        if not extended_glycan_search:
            theoretical_set = list(self.target.stub_fragments(extended=True))
        else:
            theoretical_set = list(self.target.stub_fragments(extended=True, extended_fucosylation=True))
        core_fragments = set()
        for frag in theoretical_set:
            if not frag.is_extended:
                core_fragments.add(frag.name)

        core_matches = set()
        extended_matches = set()

        for peak_pair in self.solution_map:
            if peak_pair.fragment.series != series:
                continue
            elif peak_pair.fragment_name in core_fragments:
                core_matches.add(peak_pair.fragment_name)
            else:
                extended_matches.add(peak_pair.fragment_name)
            peak = peak_pair.peak
            if peak.index.neutral_mass not in seen:
                seen.add(peak.index.neutral_mass)
        glycan_composition = self.target.glycan_composition
        n = self._get_internal_size(glycan_composition)
        m = len(theoretical_set)
        k = 2.0
        if not fragile_fucose:
            side_group_count = self._glycan_side_group_count(glycan_composition)
            if side_group_count > 0:
                k = 1.0
        d = min(max(n * math.log(n) / k, n), m)
        core_coverage = ((len(core_matches) * 1.0) /
                         len(core_fragments)) ** core_weight
        extended_coverage = min(
            float(len(core_matches) + len(extended_matches)) / d, 1.0) ** coverage_weight
        coverage = core_coverage * extended_coverage
        self._glycan_coverage = coverage
        return coverage

    def glycan_coverage(self, core_weight: float=0.4,
                        coverage_weight: float=0.5,
                        fragile_fucose:bool=False,
                        extended_glycan_search:bool=False,
                        *args, **kwargs) -> float:
        if self._glycan_coverage is not None:
            return self._glycan_coverage
        self._glycan_coverage = self._calculate_glycan_coverage(
            core_weight, coverage_weight, fragile_fucose=fragile_fucose,
            extended_glycan_search=extended_glycan_search, *args, **kwargs)
        return self._glycan_coverage

    def peptide_coverage(self) -> float:
        return self._calculate_peptide_coverage()

    def calculate_glycan_score(self, *args, **kwargs) -> float:
        return 0

    def count_peptide_Y_ion_utilization(self) -> Tuple[float, int, int, float]:
        if self._peptide_Y_ion_utilization is not None:
            return self._peptide_Y_ion_utilization
        weight = 0.0
        stub_count = 0
        peptide_backbone_count = 0
        total_utilization = 0.0
        for pfp in self.solution_map:
            frag_series = pfp.fragment.series
            if frag_series != "oxonium_ion":
                total_utilization += pfp.peak.intensity
            if frag_series == "stub_glycopeptide":
                stub_count += 1
                weight += math.log10(pfp.peak.intensity)
            elif frag_series.int_code <= IonSeries.z.int_code:
                peptide_backbone_count += 1
        self._peptide_Y_ion_utilization = weight, stub_count, peptide_backbone_count, total_utilization
        return self._peptide_Y_ion_utilization

    def oxonium_ion_utilization(self):
        return 0.0

    def get_auxiliary_data(self):
        data = super(GlycopeptideSpectrumMatcherBase, self).get_auxiliary_data()
        data['score'] = self.score
        data['glycan_score'] = self.glycan_score()
        data['peptide_score'] = self.peptide_score()
        data['glycan_coverage'] = self.glycan_coverage()
        data['n_peaks'] = len(self.spectrum)
        data['n_fragment_matches'] = len(self.solution_map)
        (data['stub_glycopeptide_intensity_utilization'],
         data['n_stub_glycopeptide_matches'],
         data['peptide_backbone_matches']) = self.count_peptide_Y_ion_utilization()
        return data

try:
    from glycresoft._c.tandem.tandem_scoring_helpers import (
        _match_oxonium_ions, _match_stub_glycopeptides,
        _calculate_glycan_coverage, _compute_glycan_coverage_from_metrics,
        _calculate_peptide_coverage, _calculate_peptide_coverage_no_glycosylated,
        count_peptide_Y_ion_utilization)
    GlycopeptideSpectrumMatcherBase._match_oxonium_ions = _match_oxonium_ions
    GlycopeptideSpectrumMatcherBase._match_stub_glycopeptides = _match_stub_glycopeptides
    GlycopeptideSpectrumMatcherBase._calculate_glycan_coverage = _calculate_glycan_coverage
    GlycopeptideSpectrumMatcherBase._compute_glycan_coverage_from_metrics = _compute_glycan_coverage_from_metrics
    GlycopeptideSpectrumMatcherBase._calculate_peptide_coverage = _calculate_peptide_coverage
    GlycopeptideSpectrumMatcherBase._calculate_peptide_coverage_no_glycosylated = _calculate_peptide_coverage_no_glycosylated
except ImportError:
    pass



import unittest

from glycresoft.structure.structure_loader import (
    FragmentCachingGlycopeptide, hashable_glycan_glycopeptide_parser,
    HashableGlycanComposition, GlycanCompositionWithOffsetProxy)

from glycopeptidepy.test.sequence_test_suite import PeptideSequenceSuiteBase

glycopeptide = "YPVLN(N-Glycosylation)VTMPN(Deamidation)NGKFDK{Hex:9; HexNAc:2}"


class TestFragmentCachingGlycopeptide(PeptideSequenceSuiteBase, unittest.TestCase):
    def parse_sequence(self, seqstr):
        return FragmentCachingGlycopeptide(seqstr)

    def test_mass(self):
        gp = FragmentCachingGlycopeptide(glycopeptide)
        self.assertAlmostEqual(gp.total_mass, 3701.5421769127897, 2)

    def test_parse(self):
        parts = hashable_glycan_glycopeptide_parser(glycopeptide)
        gc = parts[-3]
        self.assertIsInstance(
            gc, (HashableGlycanComposition, GlycanCompositionWithOffsetProxy))

    test_cad_fragmentation = None
    test_glycan_fragments_stubs = None


if __name__ == '__main__':
    unittest.main()



from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_method
from sqlalchemy.orm import relationship, backref, object_session
from sqlalchemy import (
    Column, Numeric, Integer, String, ForeignKey,
    Table, Index)

from glycresoft.serialize.base import (Base)


from glypy import Composition
from glypy.io import glycoct
from glycopeptidepy import HashableGlycanComposition

from glycresoft.serialize.hypothesis.hypothesis import GlycanHypothesis, GlycopeptideHypothesis
from glycresoft.serialize.hypothesis.generic import HasReferenceAccessionNumber, HasChemicalComposition


class GlycanBase(HasChemicalComposition):
    calculated_mass = Column(Numeric(12, 6, asdecimal=False), index=True)
    formula = Column(String(128), index=True)
    composition = Column(String(128))

    def convert(self):
        inst = HashableGlycanComposition.parse(self.composition)
        inst.id = self.id
        return inst

    _glycan_composition = None

    def __getitem__(self, key):
        if self._glycan_composition is None:
            self._glycan_composition = HashableGlycanComposition.parse(self.composition)
        return self._glycan_composition[key]

    def keys(self):
        if self._glycan_composition is None:
            self._glycan_composition = HashableGlycanComposition.parse(self.composition)
        return self._glycan_composition.keys()

    def mass(self):
        return self.calculated_mass


class GlycanComposition(GlycanBase, Base):
    __tablename__ = 'GlycanComposition'

    id = Column(Integer, primary_key=True)

    @declared_attr
    def hypothesis_id(self):
        return Column(Integer, ForeignKey(
            GlycanHypothesis.id, ondelete="CASCADE"), index=True)

    def __iter__(self):
        return iter(self.keys())

    @declared_attr
    def hypothesis(self):
        return relationship(GlycanHypothesis, backref=backref('glycans', lazy='dynamic'))

    def __repr__(self):
        return "DBGlycanComposition(%s)" % (self.composition)

    structure_classes = relationship("GlycanClass", secondary=lambda: GlycanCompositionToClass, lazy='dynamic')

    __table_args__ = (Index("ix_GlycanComposition_mass_search_index", "calculated_mass", "hypothesis_id"),)


class GlycanStructure(GlycanBase, Base, HasReferenceAccessionNumber):
    __tablename__ = "GlycanStructure"

    id = Column(Integer, primary_key=True)

    glycan_sequence = Column(String(2048), index=True)

    glycan_composition_id = Column(
        Integer, ForeignKey(GlycanComposition.id, ondelete='CASCADE'), index=True)

    glycan_composition = relationship(GlycanComposition, backref=backref("glycan_structures", lazy='dynamic'))

    @declared_attr
    def hypothesis_id(self):
        return Column(Integer, ForeignKey(
            GlycanHypothesis.id, ondelete="CASCADE"), index=True)

    @declared_attr
    def hypothesis(self):
        return relationship(GlycanHypothesis, backref=backref('glycan_structures', lazy='dynamic'))

    structure_classes = relationship("GlycanClass", secondary=lambda: GlycanStructureToClass, lazy='dynamic')

    def convert(self):
        seq = self.glycan_sequence
        structure = glycoct.loads(seq)
        structure.id = self.id
        return structure

    def __repr__(self):
        return "DBGlycanStructure:%d\n%s" % (self.id, self.glycan_sequence)

    __table_args__ = (Index("ix_GlycanStructure_mass_search_index", "calculated_mass", "hypothesis_id"),)


GlycanCombinationGlycanComposition = Table(
    "GlycanCombinationGlycanComposition", Base.metadata,
    Column("glycan_id", Integer, ForeignKey("GlycanComposition.id", ondelete="CASCADE"), index=True),
    Column("combination_id", Integer, ForeignKey("GlycanCombination.id", ondelete="CASCADE"), index=True),
    Column("count", Integer)
)


class GlycanCombination(GlycanBase, Base):
    __tablename__ = 'GlycanCombination'

    id = Column(Integer, primary_key=True)

    @declared_attr
    def hypothesis_id(self):
        return Column(Integer, ForeignKey(
            GlycopeptideHypothesis.id, ondelete="CASCADE"), index=True)

    @declared_attr
    def hypothesis(self):
        return relationship(GlycopeptideHypothesis, backref=backref('glycan_combinations', lazy='dynamic'))

    count = Column(Integer)

    components = relationship(
        GlycanComposition,
        secondary=GlycanCombinationGlycanComposition,
        lazy='dynamic')

    def __iter__(self):
        for composition, count in self.components.add_columns(
                GlycanCombinationGlycanComposition.c.count):
            i = 0
            while i < count:
                yield composition
                i += 1

    @property
    def component_ids(self):
        session = object_session(self)
        ids = session.query(GlycanCombinationGlycanComposition.c.glycan_id).filter(
            GlycanCombinationGlycanComposition.c.combination_id == self.id).all()
        return [i[0] for i in ids]

    def _get_component_classes(self):
        for case in self:
            yield case.structure_classes.all()

    _component_classes = None

    @property
    def component_classes(self):
        if self._component_classes is None:
            self._component_classes = tuple(self._get_component_classes())
        return self._component_classes

    @hybrid_method
    def dehydrated_mass(self, water_mass=Composition("H2O").mass):
        mass = self.calculated_mass
        return mass - (water_mass * self.count)

    _dehydrated_composition = None

    def dehydrated_composition(self):
        if self._dehydrated_composition is None:
            self._dehydrated_composition = self.total_composition() - (
                self.count * Composition("H2O"))
        return self._dehydrated_composition

    def __repr__(self):
        rep = "GlycanCombination({self.count}, {self.composition})".format(self=self)
        return rep


class GlycanClass(Base):
    __tablename__ = 'GlycanClass'

    id = Column(Integer, primary_key=True)
    name = Column(String(128), index=True)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return "GlycanClass(name=%r)" % (self.name,)


class _namespace(object):
    def __repr__(self):
        return "(%r)" % self.__dict__

    def __contains__(self, key):
        return key in self.__dict__

    def __getitem__(self, key):
        return self.__dict__[key]


GlycanTypes = _namespace()
GlycanTypes.n_glycan = "N-Glycan"
GlycanTypes.o_glycan = "O-Glycan"
GlycanTypes.gag_linker = "GAG-Linker"


GlycanCompositionToClass = Table(
    "GlycanCompositionToClass", Base.metadata,
    Column("glycan_id", Integer, ForeignKey("GlycanComposition.id", ondelete="CASCADE"), primary_key=True),
    Column("class_id", Integer, ForeignKey("GlycanClass.id", ondelete="CASCADE"), primary_key=True)
)


GlycanStructureToClass = Table(
    "GlycanStructureToClass", Base.metadata,
    Column("glycan_id", Integer, ForeignKey("GlycanStructure.id", ondelete="CASCADE"), primary_key=True),
    Column("class_id", Integer, ForeignKey("GlycanClass.id", ondelete="CASCADE"), primary_key=True)
)



cimport cython

cimport numpy as np

np.import_array()


@cython.freelist(100000)
cdef class SplittingPoint(object):
    cdef:
        public double first_maximum
        public double minimum
        public double second_maximum
        public double minimum_index
        public double total_distance

    @staticmethod
    cdef SplittingPoint _create(double first_maximum, double minimum, double second_maximum, double minimum_index):
        cdef:
            SplittingPoint inst
        inst = SplittingPoint.__new__(SplittingPoint)
        inst.first_maximum = first_maximum
        inst.minimum = minimum
        inst.second_maximum = second_maximum
        inst.minimum_index = minimum_index
        inst.total_distance = inst.compute_distance()
        return inst

    def __init__(self, first_maximum, minimum, second_maximum, minimum_index):
        self.first_maximum = first_maximum
        self.minimum = minimum
        self.second_maximum = second_maximum
        self.minimum_index = minimum_index
        self.total_distance = self.compute_distance()

    def __reduce__(self):
        return self.__class__, (self.first_maximum, self.minimum, self.second_maximum, self.minimum_index)

    cpdef double compute_distance(self):
        return (self.first_maximum - self.minimum) + (self.second_maximum - self.minimum)

    def __repr__(self):
        return "SplittingPoint(%0.4f, %0.4f, %0.4f, %0.2f, %0.3e)" % (
            self.first_maximum, self.minimum, self.second_maximum, self.minimum_index, self.total_distance)


@cython.binding
@cython.boundscheck(False)
def locate_extrema(self, np.ndarray[double, ndim=1, mode='c'] xs=None, np.ndarray[double, ndim=1, mode='c']ys=None):
    cdef:
        np.ndarray[np.int64_t, ndim=1, mode='c'] maxima_indices
        np.ndarray[np.int64_t, ndim=1, mode='c'] minima_indices
        list candidates
        size_t i, j, k, max_i, max_j, min_k, n_maxima, n_minima
        double y_i, y_j, y_k
        SplittingPoint point

    if xs is None:
        xs = self.xs
    if ys is None:
        ys = self.ys

    maxima_indices, minima_indices = self._extreme_indices(ys)
    candidates = []

    n_maxima = len(maxima_indices)
    n_minima = len(minima_indices)

    for i in range(n_maxima):
        max_i = maxima_indices[i]
        for j in range(i + 1, n_maxima):
            max_j = maxima_indices[j]
            for k in range(n_minima):
                min_k = minima_indices[k]
                y_i = ys[max_i]
                y_j = ys[max_j]
                y_k = ys[min_k]
                if max_i < min_k < max_j and (y_i - y_k) > (y_i * 0.01) and (
                        y_j - y_k) > (y_j * 0.01):
                    point = SplittingPoint._create(y_i, y_k, y_j, xs[min_k])
                    candidates.append(point)

    if candidates:
        best_point = max(candidates, key=lambda x: x.total_distance)
        self.partition_sites.append(best_point)

    return candidates



from collections import defaultdict

from glycresoft.database.mass_collection import NeutralMassDatabase

from glycresoft.chromatogram_tree import (
    Chromatogram,
    Unmodified,
    DuplicateNodeError,
    ChromatogramFilter,
    GlycanCompositionChromatogram,
    GlycopeptideChromatogram,
    ChromatogramGraph)

from glycresoft.task import TaskBase


def span_overlap(reference, target):
    """test whether two time Chromatogram objects
    overlap each other in the time domain

    Parameters
    ----------
    reference: Chromatogram
    target: Chromatogram

    Returns
    -------
    bool
    """
    cond = ((reference.start_time <= target.start_time and reference.end_time >= target.end_time) or (
        reference.start_time >= target.start_time and reference.end_time <= target.end_time) or (
            reference.start_time >= target.start_time and reference.end_time >= target.end_time and
            reference.start_time <= target.end_time) or (
                reference.start_time <= target.start_time and reference.end_time >= target.start_time) or (
                    reference.start_time <= target.end_time and reference.end_time >= target.end_time))
    return cond


class CompositionGroup(object):
    def __init__(self, name, members):
        self.name = name
        self.members = tuple(members)

    def __iter__(self):
        return iter(self.members)

    def __repr__(self):
        return "CompositionGroup(%r, %d members)" % (self.name, len(self.members))

    def __eq__(self, other):
        return self.members == other.members


class ChromatogramMatcher(TaskBase):
    memory_load_threshold = 1e5

    def __init__(self, database, chromatogram_type=None, require_unmodified=False, enforce_charges=False):
        if chromatogram_type is None:
            chromatogram_type = GlycanCompositionChromatogram
        self.database = database
        self.require_unmodified = require_unmodified
        self.enforce_charges = enforce_charges
        self._group_bundle = dict()
        self.chromatogram_type = chromatogram_type
        if len(database) < self.memory_load_threshold:
            self._in_memory = True
            self.database = NeutralMassDatabase(list(database.get_all_records()))
            self._original_convert = database._convert
        else:
            self._in_memory = False

    def _convert(self, record):
        if not self._in_memory:
            return self.database._convert(record)
        else:
            return self._original_convert(record)

    def _match(self, neutral_mass, mass_error_tolerance=1e-5):
        return self.database.search_mass_ppm(neutral_mass, mass_error_tolerance)

    def _prepare_group(self, key, matches):
        ids = frozenset(m.id for m in matches)
        if len(ids) == 0:
            return None
        try:
            bundle = self._group_bundle[ids]
            return bundle
        except KeyError:
            bundle = CompositionGroup(key, [
                self._convert(m)
                for m in sorted(matches, key=lambda x: getattr(x, "calculated_mass", 0))])
            self._group_bundle[ids] = bundle
            return bundle

    def match(self, mass, mass_error_tolerance=1e-5):
        hits = self._match(mass, mass_error_tolerance)
        bundle = self._prepare_group(mass, hits)
        return bundle

    def assign(self, chromatogram, group):
        out = []
        if group is None:
            return [chromatogram]
        for composition in group:
            case = chromatogram.clone(self.chromatogram_type)
            case.composition = composition
            out.append(case)
        if len(out) == 0:
            return [chromatogram]
        return out

    def search(self, chromatogram, mass_error_tolerance=1e-5):
        return self.assign(chromatogram, self.match(
            chromatogram.weighted_neutral_mass, mass_error_tolerance))

    def reverse_mass_shift_search(self, chromatograms, mass_shifts, mass_error_tolerance=1e-5):
        exclude_compositions = defaultdict(list)
        candidate_chromatograms = []

        new_members = {}
        unmatched = []

        for chroma in chromatograms:
            if chroma.composition is not None:
                exclude_compositions[chroma.composition].append(chroma)
            else:
                candidate_chromatograms.append(chroma)
        n = len(chromatograms)
        i = 0
        self.log("Begin Reverse Search")
        for chroma in candidate_chromatograms:
            i += 1
            if i % 1000 == 0:
                self.log("... %0.2f%% chromatograms searched (%d/%d)" % (i * 100. / n, i, n))
            candidate_mass = chroma.weighted_neutral_mass
            matched = False
            exclude = False
            for mass_shift in mass_shifts:
                matches = self.match(candidate_mass - mass_shift.mass, mass_error_tolerance)
                if matches is None:
                    continue
                for match in matches:
                    name = match
                    if name in exclude_compositions:
                        # This chromatogram matches another form of an existing composition
                        # assignment. If it were assigned during `join_mass_shifted`, then
                        # it overlapped with that entity and should not be merged. Otherwise
                        # construct a new match
                        for hit in exclude_compositions[name]:
                            if span_overlap(hit, chroma):
                                exclude = True
                                break
                        else:
                            if name in new_members:
                                chroma_to_update = new_members[name]
                            else:
                                chroma_to_update = self.chromatogram_type(match)
                                chroma_to_update.created_at = "reverse_mass_shiftion_search"
                            chroma, _ = chroma.bisect_mass_shift(Unmodified)
                            chroma_to_update = chroma_to_update.merge(chroma, mass_shift)
                            chroma_to_update.created_at = "reverse_mass_shiftion_search"
                            new_members[name] = chroma_to_update
                            matched = True
                    else:
                        if name in new_members:
                            chroma_to_update = new_members[name]
                        else:
                            chroma_to_update = self.chromatogram_type(match)
                            chroma_to_update.created_at = "reverse_mass_shiftion_search"
                        chroma, _ = chroma.bisect_mass_shift(Unmodified)
                        chroma_to_update = chroma_to_update.merge(chroma, mass_shift)
                        chroma_to_update.created_at = "reverse_mass_shiftion_search"
                        new_members[name] = chroma_to_update
                        matched = True
            if not matched and not exclude:
                unmatched.append(chroma)
        out = []
        out.extend(s for g in exclude_compositions.values() for s in g)
        out.extend(new_members.values())
        out.extend(unmatched)
        return ChromatogramFilter(out)

    def join_mass_shifted(self, chromatograms, mass_shifts, mass_error_tolerance=1e-5):
        out = []
        i = 0
        n = len(chromatograms)
        self.log("Begin Forward Search")
        for chroma in chromatograms:
            i += 1
            if i % 1000 == 0:
                self.log("... %0.2f%% chromatograms searched (%d/%d)" % (i * 100. / n, i, n))
            add = chroma
            for mass_shift in mass_shifts:
                query_mass = chroma.weighted_neutral_mass + mass_shift.mass
                matches = chromatograms.find_all_by_mass(query_mass, mass_error_tolerance)
                for match in matches:
                    if match and span_overlap(add, match):
                        try:
                            match.used_as_mass_shift.append((add.key, mass_shift))
                            add = add.merge(match, node_type=mass_shift, skip_duplicate_nodes=True)
                            add.created_at = "join_mass_shifted"
                            add.mass_shifts.append(mass_shift)
                        except DuplicateNodeError as e:
                            e.original = chroma
                            e.to_add = match
                            e.accumulated = add
                            e.mass_shift = mass_shift
                            raise e
            out.append(add)
        return ChromatogramFilter(out)

    def join_common_identities(self, chromatograms, delta_rt=0):
        chromatograms._build_key_map()
        key_map = chromatograms._key_map
        out = []
        for _, disjoint_set in key_map.items():
            if len(tuple(disjoint_set)) == 1:
                out.extend(disjoint_set)
                continue

            accumulated = []
            last = disjoint_set[0]
            for case in disjoint_set[1:]:
                if last.overlaps_in_time(case) or ((case.start_time - last.end_time) < delta_rt):
                    merged = last._merge_missing_only(case)
                    merged.used_as_mass_shift = list(last.used_as_mass_shift)
                    for ua in case.used_as_mass_shift:
                        if ua not in merged.used_as_mass_shift:
                            merged.used_as_mass_shift.append(ua)
                    last = merged
                    last.created_at = "join_common_identities"
                else:
                    accumulated.append(last)
                    last = case
            accumulated.append(last)
            out.extend(accumulated)
        return ChromatogramFilter(out)

    def find_related_profiles(self, chromatograms, mass_shifts, mass_error_tolerance=1e-5):
        self.log("Building Connected Components")
        graph = ChromatogramGraph(chromatograms)
        graph.find_shared_peaks()
        components = graph.connected_components()

        n_components = len(components)
        self.log("Validating %d Components" % (n_components, ))
        for i_components, component in enumerate(components):
            if i_components % 1000 == 0 and i_components > 0:
                self.log("... %d Components Validated (%0.2f%%)" % (
                    i_components,
                    i_components / float(n_components) * 100.))
            if len(component) == 1:
                continue
            component = ChromatogramFilter([node.chromatogram for node in component])

            for a in component:
                pairs = []
                for mass_shift in mass_shifts:
                    bs = component.find_all_by_mass(
                        a.weighted_neutral_mass - mass_shift.mass, mass_error_tolerance)
                    for b in bs:
                        if b != a:
                            pairs.append((mass_shift, b))
                if not pairs:
                    continue
                grouped_pairs = []
                pairs.sort(key=lambda x: (x[1].start_time, x[1].weighted_neutral_mass))
                last = [pairs[0]]
                for current in pairs[1:]:
                    if current[1] is last[0][1]:
                        last.append(current)
                    else:
                        grouped_pairs.append(last)
                        last = [current]
                grouped_pairs.append(last)
                unique_pairs = []

                def minimizer(args):
                    mass_shift, b = args
                    return abs(a.weighted_neutral_mass - (b.weighted_neutral_mass + mass_shift.mass))

                for pair_group in grouped_pairs:
                    unique_pairs.append(min(pair_group, key=minimizer))

                for mass_shift, b in unique_pairs:
                    used_set = set(b.used_as_mass_shift)
                    used_set.add((a.key, mass_shift))
                    b.used_as_mass_shift = list(used_set)

    def search_all(self, chromatograms, mass_error_tolerance=1e-5):
        matches = []
        chromatograms = ChromatogramFilter(chromatograms)
        self.log("Matching Chromatograms")
        i = 0
        n = len(chromatograms)
        for chro in chromatograms:
            i += 1
            if i % 1000 == 0 and i > 0:
                self.log("... %0.2f%% chromatograms searched (%d/%d)" % (i * 100. / n, i, n))
            matches.extend(self.search(chro, mass_error_tolerance))
        matches = ChromatogramFilter(matches)
        return matches

    def strip_only_modified(self, matches):
        out = []
        for match in matches:
            adducts = match.adducts
            if Unmodified in adducts:
                out.append(match)
        return out

    def enforce_charge_counts(self, matches):
        out = []
        for match in matches:
            changed = False
            for adduct in match.adducts:
                if adduct.charge_carrier != 0:
                    charge_carrier = adduct.charge_carrier
                    charges = match.charge_states
                    has_sufficient_charge = [c for c in charges if c >= charge_carrier]
                    has_insufficient_charge = [c for c in charges if c < charge_carrier]
                    if has_insufficient_charge:
                        # TODO: Bisect the chromatogram along the relevant charge states
                        # and then extract out the disallowed adduction state, then re-combine
                        # the chromatogram.
                        pass
            if not changed:
                out.append(match)
        return out

    def process(self, chromatograms, mass_shifts=None, mass_error_tolerance=1e-5, delta_rt=0):
        if mass_shifts is None:
            mass_shifts = []
        matches = []
        chromatograms = ChromatogramFilter(chromatograms)
        matches = self.search_all(chromatograms, mass_error_tolerance)
        matches = self.join_common_identities(matches, delta_rt)
        if mass_shifts:
            self.log("Handling Mass Shifts")
            matches = self.join_mass_shifted(matches, mass_shifts, mass_error_tolerance)
            matches = self.reverse_mass_shift_search(matches, mass_shifts, mass_error_tolerance)
        matches = self.join_common_identities(matches, delta_rt)
        self.find_related_profiles(matches, mass_shifts, mass_error_tolerance)
        if self.require_unmodified:
            matches = self.strip_only_modified(matches)
        return matches


class GlycanChromatogramMatcher(ChromatogramMatcher):
    pass


class GlycopeptideChromatogramMatcher(ChromatogramMatcher):
    def __init__(self, database, chromatogram_type=None, require_unmodified=False, enforce_charges=False):
        if chromatogram_type is None:
            chromatogram_type = GlycopeptideChromatogram
        super(GlycopeptideChromatogramMatcher, self).__init__(
            database, chromatogram_type, require_unmodified=require_unmodified,
            enforce_charges=enforce_charges)


class NonSplittingChromatogramMatcher(ChromatogramMatcher):
    def __init__(self, database, chromatogram_type=None, require_unmodified=False, enforce_charges=False):
        if chromatogram_type is None:
            chromatogram_type = Chromatogram
        super(NonSplittingChromatogramMatcher, self).__init__(
            database, chromatogram_type, require_unmodified=require_unmodified,
            enforce_charges=enforce_charges)

    def assign(self, chromatogram, group):
        out = []
        if group is None:
            return [chromatogram]
        else:
            case = chromatogram.clone(self.chromatogram_type)
            case.composition = group
            out.append(case)
            return out



import json
import warnings

from collections import defaultdict
from io import StringIO

import numpy as np

from .base import (
    UniformCountScoringModelBase,
    DecayRateCountScoringModelBase,
    LogarithmicCountScoringModelBase,
    MassScalingCountScoringModel,
    ScoringFeatureBase)


class ChargeStateDistributionScoringModelBase(ScoringFeatureBase):
    feature_type = "charge_count"

    def get_state_count(self, chromatogram):
        return chromatogram.n_charge_states

    def get_states(self, chromatogram):
        return chromatogram.charge_states

    def get_signal_proportions(self, chromatogram):
        proportions = {}
        states = self.get_states(chromatogram)
        rest = chromatogram
        total = 0
        for state in states:
            part, rest = rest.bisect_charge(state)
            proportions[state] = part.total_signal
            total += part.total_signal
        for k in proportions:
            proportions[k] /= total
        # Anything left in `rest` is from a charge state with too
        # little support to be used
        return proportions

    def reject(self, score_components, solution):
        score = score_components[self.feature_type]
        return score < 0.05


_CHARGE_MODEL = ChargeStateDistributionScoringModelBase


class UniformChargeStateScoringModel(
        _CHARGE_MODEL, UniformCountScoringModelBase):
    pass


class DecayRateChargeStateScoringModel(
        _CHARGE_MODEL, DecayRateCountScoringModelBase):
    pass


class LogarithmicChargeStateScoringModel(
        _CHARGE_MODEL, LogarithmicCountScoringModelBase):
    pass


def decay(x, step=0.4, rate=1.5):
    v = 0
    for i in range(x):
        v += (step / (i + rate))
    return v


def ones(x):
    return (x - (np.floor(x / 10.) * 10))


def neighborhood_of(x, scale=100.):
    n = x / scale
    up = ones(n) > 5
    if up:
        neighborhood = (np.floor(n / 10.) + 1) * 10
    else:
        neighborhood = (np.floor(n / 10.) + 1) * 10
    return neighborhood * scale


uniform_model = UniformChargeStateScoringModel()
decay_model = DecayRateChargeStateScoringModel()


class MassScalingChargeStateScoringModel(_CHARGE_MODEL, MassScalingCountScoringModel):
    def __init__(self, table, neighborhood_width=100., fit_information=None):
        # self.table = table
        # self.neighborhood_width = neighborhood_width
        _CHARGE_MODEL.__init__(self)
        MassScalingCountScoringModel.__init__(self, table, neighborhood_width)
        self.fit_information = fit_information or {}

    def handle_missing_neighborhood(self, chromatogram, neighborhood, *args, **kwargs):
        warnings.warn(
            ("%f was not found for this charge state "
             "scoring model. Defaulting to uniform model") % neighborhood)
        return uniform_model.score(chromatogram, *args, **kwargs)

    def handle_missing_bin(self, chromatogram, bins, key, neighborhood, *args, **kwargs):
        warnings.warn("%d not found for this mass range (%f). Using bin average (%r)" % (
            key, neighborhood, chromatogram.charge_states))
        return sum(bins.values()) / float(len(bins))

    def transform_state(self, state):
        return abs(state)

    @classmethod
    def fit(cls, observations, missing=0.01, neighborhood_width=100.,
            ignore_singly_charged=False):
        bins = defaultdict(lambda: defaultdict(float))

        fit_info = {
            "ignore_singly_charged": ignore_singly_charged,
            "missing": missing,
        }

        self = cls({}, neighborhood_width=neighborhood_width)

        for sol in observations:
            neighborhood = self.neighborhood_of(sol.neutral_mass)
            for c, val in self.get_signal_proportions(sol).items():
                c = self.transform_state(c)
                if ignore_singly_charged and c == 1:
                    continue
                bins[neighborhood][c] += 1

        model_table = {}

        all_states = set()
        for level in bins.values():
            all_states.update(level.keys())

        all_states.add(1 * (min(all_states) / abs(min(all_states))))

        for neighborhood, counts in bins.items():
            for c in all_states:
                counts[c] += missing
            total = sum(counts.values())
            entry = {k: v / total for k, v in counts.items()}
            model_table[neighborhood] = entry

        return cls(model_table, neighborhood_width, fit_information=fit_info)

    def dump(self, file_obj, include_fit_information=True):
        json.dump(
            {
                "neighborhood_width": self.neighborhood_width,
                "table": self.table,
                "fit_information": self.fit_information if include_fit_information else {}
            },
            file_obj, indent=4, sort_keys=True)

    @classmethod
    def load(cls, file_obj):
        data = json.load(file_obj)
        table = data.pop("table")
        width = float(data.pop("neighborhood_width"))

        def numeric_keys(table, dtype=float, convert_value=lambda x: x):
            return {abs(dtype(k)): convert_value(v) for k, v in table.items()}

        table = numeric_keys(table, convert_value=lambda x: numeric_keys(x, int))

        return cls(table=table, neighborhood_width=width)

    def clone(self):
        text_buffer = StringIO()
        self.dump(text_buffer)
        text_buffer.seek(0)
        return self.load(text_buffer)


class WeightedMassScalingChargeStateScoringModel(MassScalingChargeStateScoringModel):
    @classmethod
    def fit(cls, observations, missing=0.01, neighborhood_width=100.,
            ignore_singly_charged=False, smooth=0):
        bins = defaultdict(lambda: defaultdict(float))

        fit_info = {
            "ignore_singly_charged": ignore_singly_charged,
            "missing": missing,
            "smooth": smooth,
            "track": defaultdict(lambda: defaultdict(list)),
            "count": defaultdict(int)
        }

        self = cls({}, neighborhood_width=neighborhood_width)

        for sol in observations:
            neighborhood = self.neighborhood_of(sol.neutral_mass)
            fit_info['count'][neighborhood] += 1
            for c, val in self.get_signal_proportions(sol).items():
                c = self.transform_state(c)
                if ignore_singly_charged and c == 1:
                    continue
                fit_info['track'][neighborhood][c].append(val)
                bins[neighborhood][c] += val

        model_table = {}

        all_states = set()
        for level in bins.values():
            all_states.update(level.keys())

        all_states.add(1 * (min(all_states) / abs(min(all_states))))

        for neighborhood, counts in bins.items():
            largest_charge = None
            largest_charge_total = 0
            for c in all_states:
                counts[c] += missing
                if counts[c] > largest_charge_total:
                    largest_charge = c
                    largest_charge_total = counts[c]
            if smooth > 0:
                smooth_shift = largest_charge_total * smooth
                for c in all_states:
                    if c != largest_charge and counts[c] > missing:
                        counts[c] += smooth_shift

            total = sum(counts.values())
            entry = {k: v / total for k, v in counts.items()}
            model_table[neighborhood] = entry

        return cls(model_table, neighborhood_width, fit_information=fit_info)



from glycresoft.serialize.utils import temp_table
from glycresoft.serialize import (
    Peptide, Protein, DatabaseBoundOperation,
    TemplateNumberStore)

from glycresoft.task import TaskBase


class DeduplicatePeptides(DatabaseBoundOperation, TaskBase):
    def __init__(self, connection, hypothesis_id):
        DatabaseBoundOperation.__init__(self, connection)
        self.hypothesis_id = hypothesis_id

    def run(self):
        self.remove_duplicates()

    def find_best_peptides(self):
        q = self.session.query(
            Peptide.id, Peptide.peptide_score,
            Peptide.modified_peptide_sequence, Peptide.protein_id, Peptide.start_position).join(
            Protein).filter(Protein.hypothesis_id == self.hypothesis_id).yield_per(10000)
        keepers = dict()
        for id, score, modified_peptide_sequence, protein_id, start_position in q:
            key = (modified_peptide_sequence, protein_id, start_position)
            result = keepers.get(key)
            if result is None:
                keepers[key] = (id, score)
            elif result[1] < score:
                keepers[key] = (id, score)
        return keepers

    def store_best_peptides(self, keepers):
        table = temp_table(TemplateNumberStore)
        conn = self.session.connection()
        table.create(conn)
        payload = [{"value": x[0]} for x in keepers.values()]
        conn.execute(table.insert(), payload)
        self.session.commit()
        return table

    def remove_duplicates(self):
        self.log("... Deduplicating Peptides")
        keepers = self.find_best_peptides()
        self.log("... Marking Candidates")
        table = self.store_best_peptides(keepers)
        ids = self.session.query(table.c.value)
        self.log("... Building Mask")
        q = self.session.query(Peptide.id).filter(
            Peptide.protein_id == Protein.id,
            Protein.hypothesis_id == self.hypothesis_id,
            ~Peptide.id.in_(ids.correlate(None)))
        self.log("... Removing Duplicates")
        self.session.execute(Peptide.__table__.delete(
            Peptide.__table__.c.id.in_(q.selectable)))
        conn = self.session.connection()
        table.drop(conn)
        self.log("... Complete")
        self.session.commit()



import numpy as np

from .base import ScoringFeatureBase, epsilon


def total_intensity(peaks):
    return sum(p.intensity for p in peaks)


def binsearch(array, x):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        y = array[mid]
        err = y - x
        if abs(err) < 1e-4:
            return mid
        elif hi - 1 == lo:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


class TimeOffsetIndex(object):
    def __init__(self, array):
        self.array = np.array(array)
        self.average_delta = self.estimate_average_delta()

    def estimate_average_delta(self, weights=None):
        if weights is None:
            weights = np.ones(len(self) - 1)
        return np.average(self[1:] - self[:-1], weights=weights)

    def index_for(self, x):
        return binsearch(self.array, x)

    def __getitem__(self, i):
        return self.array[i]

    def __len__(self):
        return len(self.array)

    def delta(self, x):
        i = self.index_for(x)
        if i == 0:
            return self.average_delta
        try:
            y = self.array[i + 1]
            return y - x
        except IndexError:
            return self.average_delta


def blunt(x):
    if x < 0.1:
        return x
    elif 0.1 < x < 0.5:
        return np.sqrt(x) / 3.5
    else:
        return x


class ChromatogramSpacingFitter(ScoringFeatureBase):
    feature_type = "spacing_fit"

    def __init__(self, chromatogram, *args, **kwargs):
        transform_fn = kwargs.get("transform_fn")
        if transform_fn is None:
            def transform_fn(x):
                return x
        self.chromatogram = chromatogram
        self.rt_deltas = []
        self.intensity_deltas = []
        self.score = None
        self.transform_fn = transform_fn

        if len(chromatogram) < 3:
            self.score = 1.0
        else:
            self.fit()

    def transform(self, d_rt):
        return self.transform_fn(d_rt)

    def fit(self):
        times, intensities = self.chromatogram.as_arrays()
        last_rt = times[0]
        last_int = intensities[0]

        for rt, inten in zip(times[1:], intensities[1:]):
            d_rt = rt - last_rt
            self.rt_deltas.append(self.transform(d_rt))
            self.intensity_deltas.append(abs(last_int - inten))
            last_rt = rt
            last_int = inten

        self.rt_deltas = np.array(self.rt_deltas, dtype=np.float16)
        self.intensity_deltas = np.array(self.intensity_deltas, dtype=np.float32) + 1

        self.score = np.average(self.rt_deltas, weights=self.intensity_deltas)

    def __repr__(self):
        return "ChromatogramSpacingFitter(%s, %0.4f)" % (self.chromatogram, self.score)

    @classmethod
    def score(cls, chromatogram, *args, **kwargs):
        return max(1 - 2 * cls(chromatogram, *args, **kwargs).score, epsilon)


class RelativeScaleChromatogramSpacingFitter(ChromatogramSpacingFitter):

    def __init__(self, chromatogram, index, *args, **kwargs):
        self.index = index
        super(RelativeScaleChromatogramSpacingFitter, self).__init__(
            chromatogram, *args, **kwargs)

    def fit(self):
        times, intensities = self.chromatogram.as_arrays()
        last_rt = times[0]
        last_int = intensities[0]

        for rt, inten in zip(times[1:], intensities[1:]):
            d_rt = rt - last_rt
            scale = d_rt / self.index.delta(rt)
            self.rt_deltas.append(self.transform(d_rt) * scale)
            self.intensity_deltas.append(abs(last_int - inten))
            last_rt = rt
            last_int = inten

        self.rt_deltas = np.array(self.rt_deltas, dtype=np.float16)
        self.intensity_deltas = np.array(self.intensity_deltas, dtype=np.float32) + 1

        self.score = np.average(self.rt_deltas, weights=self.intensity_deltas)


class PartitionAwareRelativeScaleChromatogramSpacingFitter(RelativeScaleChromatogramSpacingFitter):
    def __init__(self, chromatogram, index, gap_size=0.25, *args, **kwargs):
        self.gap_size = gap_size
        self.partitions = [0]
        self.intensities = []
        super(PartitionAwareRelativeScaleChromatogramSpacingFitter, self).__init__(
            chromatogram, index, *args, **kwargs)

    def best_partition(self):
        i = 0
        n = len(self.partitions) - 1
        abundance = 0
        best_score = 0
        for i in range(n):
            start = self.partitions[i]
            end = self.partitions[i + 1] - 1
            score = np.average(
                self.rt_deltas[start:end],
                weights=self.intensity_deltas[start:end])
            current_abundance = np.sum(self.intensities[start:end])
            if current_abundance > abundance:
                abundance = current_abundance
                best_score = score

        return best_score

    def fit(self):
        times, intensities = self.chromatogram.as_arrays()
        last_rt = times[0]
        last_int = intensities[0]

        i = 1
        for rt, inten in zip(times[1:], intensities[1:]):
            d_rt = rt - last_rt
            if d_rt > self.gap_size:
                self.partitions.append(i)
            scale = d_rt / self.index.delta(rt)
            self.rt_deltas.append(self.transform(d_rt) * scale)
            self.intensity_deltas.append(abs(last_int - inten))
            self.intensities.append(inten)
            last_rt = rt
            last_int = inten
            i += 1

        self.partitions.append(i - 1)

        self.rt_deltas = np.array(self.rt_deltas, dtype=np.float16)
        self.intensity_deltas = np.array(self.intensity_deltas, dtype=np.float32) + 1
        self.intensities = np.array(self.intensities, dtype=np.float32)

        self.score = self.best_partition()


class ChromatogramSpacingModel(ScoringFeatureBase):
    feature_type = 'spacing_fit'

    def __init__(self, index=None, gap_size=0.25):
        self.index = index
        self.gap_size = gap_size
        self.transform_fn = None

    def configure(self, analysis_data):
        peak_loader = analysis_data['peak_loader']
        gap_size = analysis_data['delta_rt']
        self.index = TimeOffsetIndex(peak_loader.ms1_scan_times())
        tic = peak_loader.extract_total_ion_current_chromatogram()
        self.index.average_delta = self.index.estimate_average_delta(tic[1:])
        self.gap_size = gap_size
        if self.index.average_delta > 0.2:
            def transform_fn(x):
                return x / (self.index.average_delta * 15)
        else:
            transform_fn = None
        self.transform_fn = transform_fn
        return {
            "index": self.index,
            "gap_size": self.gap_size,
            "transform_fn": self.transform_fn
        }

    def fit(self, chromatogram):
        if self.index is None:
            return ChromatogramSpacingFitter(chromatogram)
        else:
            return PartitionAwareRelativeScaleChromatogramSpacingFitter(
                chromatogram, index=self.index,
                gap_size=self.gap_size, transform_fn=self.transform_fn)

    def score(self, chromatogram, *args, **kwargs):
        if self.index is None:
            return ChromatogramSpacingFitter.score(chromatogram)
        else:
            return PartitionAwareRelativeScaleChromatogramSpacingFitter.score(
                chromatogram, index=self.index, gap_size=self.gap_size,
                transform_fn=self.transform_fn)



import pkg_resources

from glycresoft.database.builder.glycan import (
    TextFileGlycanHypothesisSerializer)
from glycresoft.database.prebuilt.utils import hypothesis_register, BuildBase

from io import StringIO


def load_text():
    resource_bytes = pkg_resources.resource_string(
        __name__,
        "data/mucin_compositions.txt")
    resource_buffer = StringIO(resource_bytes.decode('utf8'))
    return resource_buffer


hypothesis_metadata = {
    "name": 'Biosynthesis Human Mucin O-Glycans',
    "hypothesis_type": 'glycan_composition',
    "description": "A collection of biosynthetically feasible mucin *O*-glycans using enzymes commonly found in humans"
    " and limited to at most 11 monosaccharides. Includes the standard cores 1-6."
}


@hypothesis_register(hypothesis_metadata['name'])
class BiosynthesisHumanMucinOGlycansBuilder(BuildBase):

    def get_hypothesis_metadata(self):
        return hypothesis_metadata

    def build(self, database_connection, **kwargs):
        if kwargs.get('hypothesis_name') is None:
            kwargs['hypothesis_name'] = (self.hypothesis_metadata['name'])
        task = TextFileGlycanHypothesisSerializer(
            load_text(), database_connection,
            **kwargs)
        task.start()
        return task



import codecs
import pkg_resources


def make_model_loader(model_type):
    def load_model(model_name):
        stream = codecs.getreader("utf-8")(
            pkg_resources.resource_stream(
                __name__, "data/%s.json" % model_name))
        return model_type.load(stream)
    return load_model



from collections import defaultdict

from multiprocessing import Manager as IPCManager

from glycopeptidepy.structure.glycan import GlycanCompositionProxy

from glycresoft.task import TaskBase
from .scoring.signature_ion_scoring import SignatureIonScorer
from ..chromatogram_mapping import ChromatogramMSMSMapper

from ..process_dispatcher import (
    IdentificationProcessDispatcher,
    SpectrumIdentificationWorkerBase)


class ChromatogramAssignmentRecord(GlycanCompositionProxy):
    def __init__(self, glycan_composition, id_key, index):
        self.glycan_composition = glycan_composition
        self.id_key = id_key
        self.index = index
        GlycanCompositionProxy.__init__(self, glycan_composition)

    @property
    def id(self):
        return self.id_key

    def __repr__(self):
        return "{self.__class__.__name__}({self.glycan_composition}, {self.id_key}, {self.index})".format(
            self=self)

    @classmethod
    def build(cls, chromatogram, default_glycan_composition=None, index=None):
        key = (chromatogram.start_time, chromatogram.end_time, chromatogram.key)
        if chromatogram.glycan_composition:
            glycan_composition = chromatogram.glycan_composition
        else:
            glycan_composition = default_glycan_composition
        return cls(glycan_composition, key, index)


class GlycanCompositionIdentificationWorker(SpectrumIdentificationWorkerBase):
    def __init__(self, input_queue, output_queue, done_event, scorer_type, evaluation_args,
                 spectrum_map, mass_shift_map, log_handler, solution_packer):
        SpectrumIdentificationWorkerBase.__init__(
            self, input_queue, output_queue, done_event, scorer_type, evaluation_args,
            spectrum_map, mass_shift_map, log_handler=log_handler, solution_packer=solution_packer)

    def evaluate(self, scan, structure, *args, **kwargs):
        target = structure
        matcher = self.scorer_type.evaluate(scan, target, *args, **kwargs)
        return matcher


class SignatureIonMapper(TaskBase):

    # simple default value from experimentation
    minimum_score = 0.05

    def __init__(self, tandem_scans, chromatograms, scan_id_to_rt=lambda x: x,
                 mass_shifts=None, minimum_mass=500, batch_size=1000,
                 default_glycan_composition=None, scorer_type=None,
                 n_processes=4):
        if scorer_type is None:
            scorer_type = SignatureIonScorer
        if mass_shifts is None:
            mass_shifts = []
        self.chromatograms = chromatograms
        self.tandem_scans = sorted(
            tandem_scans, key=lambda x: x.precursor_information.extracted_neutral_mass,
            reverse=True)
        self.scan_id_to_rt = scan_id_to_rt
        self.mass_shifts = mass_shifts
        self.minimum_mass = minimum_mass
        self.default_glycan_composition = default_glycan_composition
        self.default_glycan_composition.id = -1
        self.scorer_type = scorer_type
        self.n_processes = n_processes
        self.ipc_manager = IPCManager()

    def prepare_scan_set(self, scan_set):
        if hasattr(scan_set[0], 'convert'):
            out = []
            # Account for cases where the scan may be mentioned in the index, but
            # not actually present in the MS data
            for o in scan_set:
                try:
                    scan = (self.scorer_type.load_peaks(o))
                    out.append(scan)
                except KeyError:
                    self.log("Missing Scan: %s" % (o.id,))
            scan_set = out
        out = []
        for scan in scan_set:
            try:
                out.append(scan)
            except AttributeError as e:
                self.log("Missing Scan: %s %r" % (scan.id, e))
                continue
        return out

    def map_to_chromatograms(self, precursor_error_tolerance=1e-5):
        mapper = ChromatogramMSMSMapper(
            self.chromatograms, error_tolerance=precursor_error_tolerance,
            scan_id_to_rt=self.scan_id_to_rt)
        for scan in self.tandem_scans:
            scan_time = mapper.scan_id_to_rt(scan.precursor_information.precursor_scan_id)
            hits = mapper.find_chromatogram_spanning(scan_time)
            if hits is None:
                continue
            match = hits.find_all_by_mass(
                scan.precursor_information.neutral_mass,
                precursor_error_tolerance)
            if match:
                for m in match:
                    m.add_solution(scan)
            for mass_shift in self.mass_shifts:
                match = hits.find_all_by_mass(
                    scan.precursor_information.neutral_mass - mass_shift.mass,
                    precursor_error_tolerance)
                if match:
                    for m in match:
                        m.add_solution(scan)
        return mapper

    def _build_scan_to_entity_map(self, annotated_chromatograms):

        scan_map = dict()
        hit_map = dict()
        hit_to_scan = dict()
        default = self.default_glycan_composition
        for i, chroma in enumerate(annotated_chromatograms, 1):
            if not chroma.tandem_solutions:
                continue
            default = default.clone()
            default.id = -i
            record = ChromatogramAssignmentRecord.build(chroma, default, index=i - 1)
            hit_map[record.id] = record
            scans = []
            for scan in self.prepare_scan_set(chroma.tandem_solutions):
                scan_map[scan.id] = scan
                scans.append(scan.id)
            hit_to_scan[record.id] = scans
        return scan_map, hit_map, hit_to_scan

    def _chunk_chromatograms(self, chromatograms, batch_size=3500):
        chunk = []
        k = 0
        for chroma in chromatograms:
            chunk.append(chroma)
            k += len(chroma.tandem_solutions)
            if k >= batch_size:
                yield chunk
                chunk = []
                k = 0
        yield chunk

    def _score_mapped_tandem_parallel(self, annotated_chromatograms, *args, **kwargs):
        for chunk in self._chunk_chromatograms(annotated_chromatograms):
            scan_map, hit_map, hit_to_scan = self._build_scan_to_entity_map(chunk)
            ipd = IdentificationProcessDispatcher(
                worker_type=GlycanCompositionIdentificationWorker,
                scorer_type=self.scorer_type,
                init_args={}, evaluation_args=kwargs,
                n_processes=self.n_processes,
                ipc_manager=self.ipc_manager)
            scan_solution_map = ipd.process(scan_map, hit_map, hit_to_scan, defaultdict(lambda: "Unmodified"))

            for scan_id, matches in scan_solution_map.items():
                for match in matches:
                    chroma = chunk[match.target.index]
                    updated_scans = []
                    for scan in chroma.tandem_solutions:
                        if scan.scan_id == scan_id:
                            # match.scan = SpectrumReference(scan_id, scan.precursor_information)
                            match.target = match.target.glycan_composition
                            updated_scans.append(match)
                        else:
                            updated_scans.append(scan)
                    chroma.tandem_solutions = updated_scans
        return annotated_chromatograms

    def _score_mapped_tandem_sequential(self, annotated_chromatograms, *args, **kwargs):
        i = 0

        ni = len(annotated_chromatograms)
        for chroma in annotated_chromatograms:
            i += 1
            if i % 500 == 0:
                self.log("... Handling chromatogram %d/%d (%0.3f%%)" % (i, ni, (i * 100. / ni)))
            tandem_scans = self.prepare_scan_set(chroma.tandem_solutions)
            if chroma.glycan_composition is None:
                if self.default_glycan_composition is None:
                    continue
                else:
                    glycan_composition = self.default_glycan_composition
            else:
                glycan_composition = chroma.glycan_composition
            solutions = []
            j = 0
            nj = len(tandem_scans)
            for scan in tandem_scans:
                j += 1
                if j % 500 == 0:
                    self.log("...... Handling spectrum match %d/%d (%0.3f%%)" % (j, nj, (j * 100. / nj)))

                solution = self.scorer_type.evaluate(
                    scan, glycan_composition,
                    *args, **kwargs)
                solutions.append(solution)
            chroma.tandem_solutions = solutions
        return annotated_chromatograms

    def score_mapped_tandem(self, annotated_chromatograms, *args, **kwargs):
        if self.n_processes > 1:
            annotated_chromatograms = self._score_mapped_tandem_parallel(
                annotated_chromatograms, **kwargs)
        else:
            annotated_chromatograms = self._score_mapped_tandem_sequential(
                annotated_chromatograms, **kwargs)
        return annotated_chromatograms
